# -*- coding: utf-8 -*-

from typing import List, Union
import pandas as pd
import math
import numpy as np
import inspect

from ansys.aedt.core import Desktop, Hfss, Q3d
from ansys.aedt.core.modeler.cad.primitives import Polyline

from qiskit_metal.renderers.renderer_base import QRendererAnalysis

#The below imports are for typecheck and will probably be removed if move to the open side.
from qiskit_metal.designs import QDesign
from qiskit_metal.toolbox_metal.bounds_for_path_and_poly_tables import BoundsForPathAndPolyTables
from qiskit_metal.toolbox_metal.parsing import parse_entry
from qiskit_metal.draw.utility import to_vec3D_list
from qiskit_metal import Dict

from qiskit_metal import config
if not config.is_building_docs():
    from qiskit_metal.toolbox_python.utility_functions import get_clean_name
    from qiskit_metal.toolbox_python.utility_functions import good_fillet_idxs


class QPyaedt(QRendererAnalysis):
    """  Will fill this out later.
    """

    # Get name of subclass to set project and design name
    GROUPS = ["Non Model", "Solids", "Unclassified", "Sheets", "Lines"]

    renderer_types = ['HFSS_DM', 'HFSS_EM', 'Q3D']

    # yapf: disable
    default_options = Dict(
        begin_disable_autosave = True, # If True, Ansys will execute faster.
        close_enable_autosave = True, # Before Ansys is closed, enable autosave.

        x_buffer_width_mm=0.25,  # Buffer between max/min x and edge of ground plane, in mm
        y_buffer_width_mm=0.25,  # Buffer between max/min y and edge of ground plane, in mm

        # For HFSS
        sample_holder_top='1.2mm',  # how tall is the vacuum above center_z
        sample_holder_bottom='-1.2mm',  # how tall is the vacuum below z=0

        Lj=10e-9,  # Lj has been previously with units of nanoHenries (nH), pyaedt expects (H)
        #Lj='10nH',  # Lj has units of nanoHenries (nH)
        Cj=0  # Cj *must* be 0 for pyEPR analysis! Cj has units of femtofarads (fF)
        # _Rj=0,  # _Rj *must* be 0 for pyEPR analysis! _Rj has units of Ohms
        # project_path=None,  # default project path; if None --> get active
        # project_name=project_name,  # default project name
        # design_name=design_name,  # default design name

        # bounding_box_scale_x = 1.2, # Ratio of 'main' chip width to bounding box width
        # bounding_box_scale_y = 1.2, # Ratio of 'main' chip length to bounding box length


        # wb_threshold = '400um',
        # wb_offset = '0um',
        # wb_size = 5,
        # plot_ansys_fields_options = Dict(
        #     name="NAME:Mag_E1",
        #     UserSpecifyName='0',
        #     UserSpecifyFolder='0',
        #     QuantityName= "Mag_E",
        #     PlotFolder= "E Field",
        #     StreamlinePlot= "False",
        #     AdjacentSidePlot= "False",
        #     FullModelPlot= "False",
        #     IntrinsicVar= "Phase=\'0deg\'",
        #     PlotGeomInfo_0= "1",
        #     PlotGeomInfo_1= "Surface",
        #     PlotGeomInfo_2= "FacesList",
        #     PlotGeomInfo_3= "1",
        # ),
    )
    """Default options"""
    # yapf: enable

    NAME_DELIM = r"_"
    """Name delimiter"""

    # This is expected to be populated by each child/grandchild of the renderer.
    # It will not be aggregated, just overwritten.
    default_setup = Dict()
    """Default setup."""

    # When additional columns are added to QGeometry, this is the example to populate it.
    # e.g. element_extensions = dict(
    #         base=dict(color=str, klayer=int),
    #         path=dict(thickness=float, material=str, perfectE=bool),
    #         poly=dict(thickness=float, material=str), )
    """Element extensions dictionary   element_extensions = dict() from base class"""

    # Add columns to junction table during QAnsysRenderer.load()
    # element_extensions  is now being populated as part of load().
    # Determined from element_table_data.

    # Dict structure MUST be same as  element_extensions!!!!!!
    # This dict will be used to update QDesign during init of renderer.
    # Keeping this as a cls dict so could be edited before renderer is instantiated.
    # To update component.options junction table.

    element_table_data = dict(
        path=dict(wire_bonds=False),
        junction=dict(inductance=default_options["Lj"],
                      capacitance=default_options["Cj"]),
        #resistance=default_options["_Rj"]),
    )
    """Element table data."""

    @classmethod
    def _get_child_class_name(cls) -> str:
        parents = inspect.getmro(cls)
        return parents[0].__name__

    def __init__(self,
                 design: "QDesign",
                 renderer_type: str,
                 project_name: Union[str, None] = None,
                 design_name: Union[str, None] = None,
                 initiate=False,
                 options: Dict = None):
        """_summary_

        Args:
            design (QDesign): Use QGeometry within either QDesign or MultiPlanar design to
                            obtain elements for Ansys.
            renderer_type (str): Choose string from list ['HFSS_DM', 'HFSS_EM', 'Q3D'] for the type of
                                design to insert into project.
            project_name (Union[str, None], optional):  Give a name, or one will be made based on class name of renderer.
                                            Defaults to None.
            design_name (Union[str, None], optional):Give a name, or one will be made based on class name of renderer.
                                            Defaults to None.
            initiate (bool, optional):  True to initiate the renderer. Defaults to False.
            options (Dict, optional): Used to override all options. Defaults to None.
        """
        self._desktop = None
        # Initialize renderer first so we can use self.options

        # if initiate is true then Ansys will be at self._desktop.
        super().__init__(design=design, initiate=initiate, options=options)

        # Keep this for self awareness.
        self.renderer_type = renderer_type.upper()

        self.current_app = None
        self.project_name = project_name
        self.design_name = design_name

        # _desktop only created if initiate=True
        if self._desktop:
            self.populate_project_and_design()

        self.box_plus_buffer = False  # For now, just use chip size.

        # Note, at this point, self._desktop should have pyaedt.

        # Default behavior is to render all components unless a strict subset was chosen
        self.render_everything = True
        self.qcomp_ids = []
        self.case = -1

        self.box_for_ansys = None
        self.chip_names_matched = None  # bool
        self.valid_chip_names = None  # set of valid chip names from layer_stack
        self.chip_subtract_dict = Dict()

        # For the components which were selected to render,
        # hold both path and poly dataframes concatenated.
        self.path_and_poly_with_valid_comps = None
        # hold both path, poly and dataframes concatenated.
        self.path_poly_and_junction_with_valid_comps = None

        self.bounds_handler = BoundsForPathAndPolyTables(self.design)
        # Make box for each row that has fill==True.
        self.fill_info = dict()
        # self.fill_info will hold the name of newly generated box,
        # along with information from layer stack
        # Moved to each render_design since this method was causing error when trying to make columns of qgeomtry tables.
        #self.fill_info = self.design.ls.get_layer_datatype_when_fill_is_true()

        self.open_pins_is_valid = None

        # Before opening the application, activate Pandas before opening Ansys.
        from ansys.aedt.core import settings
        settings.enable_pandas_output = True

    def initialized(self):
        """Abstract method. Must be implemented by the subclass.
        Is renderer ready to be used?
        Implementation must return boolean True if successful. False otherwise.
        """
        return True

    # pylint: disable=arguments-renamed
    def render_component(self, component):
        """Abstract method. Must be implemented by the subclass.
        Render the specified component.

        Args:
            component (QComponent): Component to render.
        """
        pass

    def render_chips(self,
                     draw_sample_holder: bool = True,
                     box_plus_buffer: bool = True):
        """Render all chips containing components in self.qcomp_ids.

        Args:
            draw_sample_holder (bool, optional): Option to draw vacuum box around chip. Defaults to True.
            box_plus_buffer (bool, optional): Whether or not to use a box plus buffer. Defaults to True.
        """
        # Abstract method. Must be implemented by the subclass.
        # Render all chips of the design.
        # Calls render_chip for each chip.

        a = 5
        pass

        chip_list = self.get_chip_names()
        for chip_name in chip_list:

            # This is done obtained at the start of render_design
            #  since the box includes all the chips and placed in self.box_for_ansys
            #  We don't do this for just one chip, as was done for planar.
            # if box_plus_buffer:
            #     # Get bounding box of components first
            #     pass
            # else:
            #     # Adhere to chip placement and dimensions in QDesign
            #     pass
            self.render_chip(chip_name, draw_sample_holder)

            if draw_sample_holder:  # HFSS
                # Fill this out when working on HFSS
                pass

    def render_chip(self, chip_name: str, draw_sample_holder: bool):
        """
        Render individual chips.

        Args:
            chip_name (str): Name of chip.
            draw_sample_holder (bool): Option to draw vacuum box around chip.
        """

        # This is for HFSS, will do later.  Also, we may want to do just once, BUT
        # not for every chip.
        a = 5
        pass

    def render_components(self, table_type: str, layer_num: int,
                          port_list: Union[list, None], jj_to_port: Union[list,
                                                                          None],
                          ignored_jjs: Union[list, None]):
        """
        Render components by breaking them down into individual elements.
        Render all components of the design.
        If selection is none, then render all components.

        Args:
            table_type (str): Table type (poly, path, or junction).
        """
        table = self.design.qgeometry.tables[table_type]

        #If user selected a subset, then need to use subset of table,
        #otherwise use the whole table.
        if self.qcomp_ids:
            mask_component = table['component'].isin(self.qcomp_ids)
            table = table[mask_component]

        mask_layer = table['layer'] == layer_num
        table_to_use = table[mask_layer]

        for _, qgeom in table_to_use.iterrows():
            self.render_element(qgeom,
                                bool(table_type == 'junction'),
                                port_list=port_list,
                                jj_to_port=jj_to_port,
                                ignored_jjs=ignored_jjs)

        # if table_type == 'path':
        #     self.auto_wirebonds(table)

    def render_element(self, qgeom: pd.Series, is_junction: bool,
                       port_list: Union[list, None],
                       jj_to_port: Union[list, None], ignored_jjs: Union[list,
                                                                         None]):
        #Use the method in the child class.
        pass

    # pylint: disable=arguments-renamed
    def render_element_path(self, qgeom: pd.Series):
        """Render one row from the qgeometry table.

        Args:
            qgeom (pd.Series): One row from the qgeometry table.
        """
        #super().render_element_path(qgeom)

        qc_name = self.design._components[qgeom["component"]].name
        qc_elt = get_clean_name(qgeom["name"])

        qc_shapely = qgeom.geometry  # shapely geom
        subtract = qgeom['subtract']  # expect a bool
        name = f"{qc_elt}{QPyaedt.NAME_DELIM}{qc_name}"

        qc_width = parse_entry(qgeom.width)

        layer = int(qgeom.layer)
        datatype = 0
        chip_name = str(qgeom.chip)
        result = self.design.ls.get_properties_for_layer_datatype(
            ['thickness', 'z_coord', 'material', 'fill'], layer, datatype)
        if result:
            thickness, z_coord, material, fill_value = result
        else:
            self.design.ls.layer_stack_handler_pilot_error()

        # LINESTRING does not have interior and exterior coords.
        points_2d = list(qc_shapely.coords)
        points_3d = to_vec3D_list(points_2d, z_coord + (thickness / 2))

        a_polyline = self.current_app.modeler.create_polyline(
            points_3d,
            name=name,
            cover_surface=False,
            close_surface=False,
            matname=material)

        # The transparency is hardcoded.  Ugh.
        # However, it was hardcoded without pyaedt, so for now, will leave it alone.
        self.current_app.modeler[name].transparency = 0.0

        # Note: Linestring does not have interior and exterior coords.
        qc_fillet = round(qgeom.fillet, 7)
        if qc_fillet > 0:
            self.add_fillet_linestring(qgeom, points_3d, qc_fillet, a_polyline)

        if subtract and not fill_value:
            self.logger.warning(
                '\nsubtract==True and Fill is False'
                '\nWe don\'t render the geometry since there is nothing to subtract from.'
                'FYI: The fill value comes from layer_stack, '
                'and the subtract value comes from component.')
            return

        if qc_width > 0:
            a_polyline_wider = self.add_linestring_width(qc_width,
                                                         a_polyline,
                                                         points_3d,
                                                         material=material)
            a_name = a_polyline_wider.name
            a_polyline_width_thickened = self.current_app.modeler.thicken_sheet(
                a_polyline_wider.id, thickness=thickness, bBothSides=True)

        else:
            # no widening, so just thicken it.
            a_polyline_width_thickened = self.current_app.modeler.thicken_sheet(
                a_polyline.id, thickness=thickness, bBothSides=True)

        # # Thicken after fillet.
        # self.q3d.modeler.thicken_sheet(a_polyline.id, thickness=thickness)

        if subtract is True and fill_value is True:
            # When doing subtract, the layer is the key to get information
            # Layer numbers MUST be unique in the layer_stack file.
            # If box_plus_buffer is False,
            # to determine ground size, chip names are gotten from layer_stack file.

            # Need to group subtract geometries by layer.
            if layer not in self.chip_subtract_dict:
                self.chip_subtract_dict[layer] = list()

            self.chip_subtract_dict[layer].append(a_polyline_width_thickened.id)

            # self.chip_subtract_dict[layer].append(
            #     self.q3d.modeler.object_id_dict[name])

    # pylint: disable=arguments-renamed
    def render_element_poly(self, qgeom: pd.Series):
        """Render a closed polygon.

        Args:
            qgeom (pd.Series): GeoSeries of element properties.
        """
        #super().render_element_poly(qgeom)
        ansys_options = dict(transparency=0.0)
        # pylint: disable=protected-access
        qc_name = self.design._components[qgeom["component"]].name
        qc_elt = get_clean_name(qgeom["name"])

        qc_shapely = qgeom.geometry  # shapely geom
        layer = int(qgeom.layer)
        datatype = 0
        chip_name = str(qgeom.chip)
        subtract = qgeom['subtract']  # expect a bool
        # qgeom.helper is ignored since the material for each layer is
        # obtained from the layer_stack.

        result = self.design.ls.get_properties_for_layer_datatype(
            ['thickness', 'z_coord', 'material', 'fill'], layer, datatype)

        if result:
            thickness, z_coord, material, fill_value = result
        else:
            self.design.ls.layer_stack_handler_pilot_error()

        if subtract and not fill_value:
            self.logger.warning(
                '\nsubtract==True and Fill is False'
                '\nWe don\'t render the geometry since there is nothing to subtract from.'
                'FYI: The fill value comes from layer_stack, '
                'and the subtract value comes from component.')
            return

        #still need to deal with fillet for polygons?????
        qc_fillet = round(qgeom.fillet, 7)

        name = f"{qc_elt}{QPyaedt.NAME_DELIM}{qc_name}"

        points_3d = to_vec3D_list(list(qc_shapely.exterior.coords),
                                  z_coord + (thickness / 2))

        a_polyline = self.current_app.modeler.create_polyline(
            points_3d,
            name=name,
            cover_surface=True,
            close_surface=True,
            matname=material)

        # Subtract interior shapes, if any
        inner_shape_ids = []
        if len(qc_shapely.interiors) > 0:
            for i, x in enumerate(qc_shapely.interiors):
                points_3d_interior = to_vec3D_list(list(x.coords),
                                                   z_coord + (thickness / 2))
                inner_shape = self.current_app.modeler.create_polyline(
                    points_3d_interior[:-1],
                    close_surface=True,
                    cover_surface=True,
                    matname=material)
                inner_shape_ids.append(inner_shape.id)
            result = self.current_app.modeler.subtract(name,
                                                       inner_shape_ids,
                                                       keepOriginals=False)
            self.current_app.modeler.thicken_sheet(name,
                                                   thickness=thickness,
                                                   bBothSides=True)
        else:  # No interiors

            self.current_app.modeler.thicken_sheet(a_polyline.id,
                                                   thickness=thickness,
                                                   bBothSides=True)

        if subtract is True and fill_value is True:
            # Need to group subtract geometries by layer from layer stack.
            # All names of geometries with different datatypes within same layer
            # will be grouped together.

            if layer not in self.chip_subtract_dict:
                self.chip_subtract_dict[layer] = list()

            self.chip_subtract_dict[layer].append(
                self.current_app.modeler.object_id_dict[name])

            # # Input chip info into self.chip_subtract_dict
            # if chip_name not in self.chip_subtract_dict:
            #     self.chip_subtract_dict[chip_name] = Dict()

            # # Need to group subtract geometries by chip and layer using z_coord from layer stack.
            # if layer not in self.chip_subtract_dict[chip_name][layer]:
            #     self.chip_subtract_dict[chip_name][layer] = set()

            # self.chip_subtract_dict[chip_name][layer].add(name)

        a = 5

    def _initiate_renderer(self):
        """
        Open a session of the default Ansys EDT.
        Establishes the connection to the App and Desktop only.
        """
        # https://aedtdocs.pyansys.com/API/_autosummary/pyaedt.Desktop.html
        # Don't need to add the version: Version of AEDT to use.
        # The default is None, in which case the active setup or latest installed version is used.
        desktop = Desktop(version=None,
                          non_graphical=False,
                          new_desktop=True,
                          close_on_exit=False,
                          student_version=False)

        self._desktop = desktop
        if self.options.begin_enable_autosave:
            self._desktop.disable_autosave()

        # with Desktop(specified_version="2021.2",
        #              non_graphical=False,
        #              new_desktop_session=True,
        #              close_on_exit=False,
        #              student_version=False) as desktop:

        #     self._desktop = desktop

        #     # Desktop is automatically released here.

    def force_exit_ansys(self):
        """This will allow user to force closing of Ansys.
        """

        self._desktop.force_close_desktop()

    def _close_renderer(self):
        """
        Step required to close the renderer after final execution.
        Close the connection to Ansys through pyaedt. Implementation must
        return boolean True if successful. False otherwise.

        Returns:
            bool: True
        """
        #Can use any one of three. In future, we may want to switch.
        #self._desktop.close_desktop
        if self.options.close_enable_autosave:
            self._desktop.enable_autosave()
        self._desktop.release_desktop(close_projects=False, close_on_exit=False)
        #self._desktop.force_close_desktop()
        #self._desktop.close_on_exit

        #Could add more logic for returning False.
        return True

    def disconnect_ansys(self):
        """Disconnect Ansys. In future, may want this to behave differently."""
        if self._desktop:
            self._close_renderer()

    def close(self):
        """Alias of _close_renderer()

        Returns:
            bool: True
        """
        return self._close_renderer()

    def save_screenshot(self, path: str = None, show: bool = True):
        """Save the screenshot.

        Args:
            path (str, optional): Path to save location.  Defaults to None.
            show (bool, optional): Whether or not to display the screenshot.  Defaults to True.

        Returns:
            pathlib.WindowsPath: path to png formatted screenshot.
        """
        pass

    # pylint: disable=arguments-differ
    def render_design(self,
                      selection: Union[List, None] = None,
                      box_plus_buffer: bool = True):
        """Must be implemented by the subclass to finish the logic for HFSS  OR Q3D within project.

        Renders all design chips and components.
        Note: This needs to be extended with additional logic for Q3D and HFSS.
        """
        # pylint: disable=attribute-defined-outside-init
        # # They are reset for each time render_design happens.
        self.box_plus_buffer = box_plus_buffer

        # Activate project_name and design_name before anything else
        self.activate_user_project_design()
        self.clean_user_design()

        self.chip_subtract_dict.clear()
        # Get from layer_stack, previously everything was a perfect E, except for junction., can't apply perfect boundary to 3d.
        self.assign_perfE = []
        self.assign_mesh = []
        self.qcomp_ids, self.case = self.get_unique_component_ids(selection)
        self.path_and_poly_with_valid_comps = None

        # Add the buffer, using options for renderer.
        x_buff = parse_entry(self._options["x_buffer_width_mm"])
        y_buff = parse_entry(self._options["y_buffer_width_mm"])

        result = self.bounds_handler.get_bounds_of_path_and_poly_tables(
            box_plus_buffer, self.qcomp_ids, self.case, x_buff, y_buff)

        self.box_for_ansys, self.path_and_poly_with_valid_comps, self.path_poly_and_junction_with_valid_comps, self.chip_names_matched, self.valid_chip_names = result

        # Possibly this could be done once, but if ALL, the geometries are all cleared within active design,
        # these boxes would need to be generated. We can move this to inti when clear_active_design is customized.
        self.create_fill_true_box()

    def create_fill_true_box(self):
        """ Create 3D box used for both fill and or as a
        subtract box for geometries when subtract==True from qgeometry table.
        """
        (minx, miny, maxx, maxy) = self.box_for_ansys

        for key, value in self.fill_info.items():
            # Don't need to close the polygon. But for now, being consistent with rendering Poly.
            fill_rect = [
                [minx, miny, value['z_coord'] + (value['thickness'] / 2)],
                [maxx, miny, value['z_coord'] + (value['thickness'] / 2)],
                [maxx, maxy, value['z_coord'] + (value['thickness'] / 2)],
                [minx, maxy, value['z_coord'] + (value['thickness'] / 2)],
                [minx, miny, value['z_coord'] + (value['thickness'] / 2)]
            ]

            box_name = f'layer_{value["layer"]}_datatype_{value["datatype"]}_plane'

            box_at_layer_datatype = self.current_app.modeler.create_polyline(
                fill_rect,
                name=box_name,
                cover_surface=True,
                close_surface=True,
                matname=value['material'])

            self.current_app.modeler.thicken_sheet(box_at_layer_datatype.id,
                                                   thickness=value['thickness'],
                                                   bBothSides=True)

            self.fill_info[key]['box_id'] = box_at_layer_datatype.id
            self.fill_info[key]['box_name'] = box_at_layer_datatype.name

    def aedt_render_by_layer_then_tables(self,
                                         open_pins: Union[list, None],
                                         port_list: Union[list, None],
                                         jj_to_port: Union[list, None],
                                         ignored_jjs: Union[list, None],
                                         skip_junction: bool = False,
                                         data_type: int = 0):
        pass

    def add_linestring_width(self, qc_width: float, a_polyline: Polyline,
                             points_3d: list, material: str) -> Polyline:
        """Determine the orthogonal vector.  Then sweep along a_polyline using qc_width.
        Then return the reference to new polyline with sweep in Ansys.

        Args:
            qc_width (float): The width of polyline to generate.
            a_polyline (Polyline): The center of desired polyline.
            points_3d (list): [[x0,y0,z0], ....[xn,yn,zn]] List of list of points used for a_polyline.
            material (str):  The material to widen.

        Returns:
            Polyline: Reference to polyline that was widened.
        """

        x0, y0, z0 = points_3d[0]
        x1, y1, z1 = points_3d[1]
        vlen = math.sqrt((x1 - x0)**2 + (y1 - y0)**2 + (z1 - z0)**2)

        # yapf: disable
        p0 = np.array([x0, y0, z0]) + qc_width / (2 * vlen) * np.array([y0 - y1, x1 - x0, 0])
        p1 = np.array([x0, y0, z0]) + qc_width / (2 * vlen) * np.array([y1 - y0, x0 - x1, 0])
        # yapf: enable

        shortline_name = f'{a_polyline.name}_shortline'
        shortline = self.current_app.modeler.create_polyline(
            [p0, p1],
            name=shortline_name,
            close_surface=False,
            matname=material)

        a_polyline_width = shortline.sweep_along_path(a_polyline)
        #a_polyline_width = a_polyline.sweep_along_path(shortline_name)
        return a_polyline_width

    def add_one_endcap(self,
                       comp_name: str,
                       pin_name: str,
                       layer: int,
                       endcap_str: str = 'endcap'):
        """ Create a rectangle, then add thickness and denote to be
        subtracted for open pin.

        Args:
            comp_name (str): Component which has the open pin.
            pin_name (str): Name of pin which in open within the component.
            layer (int): The layer which the user choose to put the component on.
            endcap_str (str): Used to name the geometry when rendered to Ansys.
        """
        pin_dict = self.design.components[comp_name].pins[pin_name]
        width, gap = parse_entry([pin_dict["width"], pin_dict["gap"]])
        mid, normal = parse_entry(pin_dict["middle"]), pin_dict["normal"]
        chip_name = self.design.components[comp_name].options.chip
        result = self.design.ls.get_properties_for_layer_datatype(
            ['thickness', 'z_coord', 'material', 'fill'], layer)
        if result:
            thickness, z_coord, material, fill = result
        else:
            self.design.ls.layer_stack_handler_pilot_error()

        endcap_name = f"{endcap_str}_{comp_name}_{pin_name}_{layer}"

        start_vec_xyz = mid.tolist()
        start_vec_xyz.append(z_coord + (thickness / 2))

        end_vec_xyz = (mid + gap * normal).tolist()
        end_vec_xyz.append(z_coord + (thickness / 2))

        #Give a vector in direction of normal and add a width.
        endcap_reference = self.current_app.modeler.create_polyline(
            [start_vec_xyz, end_vec_xyz],
            name=endcap_name,
            cover_surface=False,
            matname=material,
            close_surface=False,
            xsection_type='Line',
            xsection_width=2 * gap + width)

        thickened = self.current_app.modeler.thicken_sheet(endcap_reference.id,
                                                           thickness=thickness,
                                                           bBothSides=True)
        if fill:
            if layer not in self.chip_subtract_dict:
                self.chip_subtract_dict[layer] = list()

            self.chip_subtract_dict[layer].append(thickened.id)

    def add_endcaps(self,
                    layer_open_pins_list: list,
                    layer_num: int,
                    endcap_str: str = 'endcap'):
        """Add a list of end-caps for open pins for a specific layer ONLY.
        This method ASSUMES there is something in open_pins list.
        This method ASSUMES that the components and corresponding pins
        are within design.
        This method ASSUMES the list already corresponds to the layer number.
        This method wil make a sublist of open_pins for each the layer_num passed.


        Args:
            open_pins (list): list of tuples with (component name , pin name)  JUST
                            for the layer passed to this method.
            layer_num (int): The layer which the end-caps should be on.
            endcap_str (str): Used to name the geometry when rendered to Ansys.
        """
        if layer_open_pins_list:
            # Now add the end-caps for this layer only!
            for comp_name, pin_name in layer_open_pins_list:
                self.add_one_endcap(comp_name, pin_name, layer_num, endcap_str)

    def add_fillet_linestring(self, qgeom: pd.Series, points_3d: list,
                              qc_fillet: float, a_polyline: Polyline):
        """Determine the idx of Polyline vertices to fillet and fillet them.

        Args:
            qgeom (pd.Series): One row from QGeometry table._
            points_3d (list): Each entry is list of x,y,z vertex.
            qc_fillet (float): Radius of fillet value
            a_polyline (Polyline): A pyaetd primitive
                                                        used for linestring from qgeometry table.
        """
        qc_fillet = parse_entry(qc_fillet)
        idxs_to_fillet = good_fillet_idxs(
            points_3d,
            qc_fillet,
            precision=self.design._template_options.PRECISION,
            isclosed=False,
        )
        if idxs_to_fillet:
            vertices = a_polyline.vertices

            for idx in idxs_to_fillet:
                vertices[idx].fillet(radius=qc_fillet)

    def subtract_from_ground(self, layer_num: int, data_type: int = 0):
        """For each chip, subtract all "negative" shapes residing on its
        surface if any such shapes exist.
        Data is stored by layer, but we always assume each layer to be unique

        Need to use a rectangular for each layer, which has fill,
        has been made as part of create_fill_true_box.
        """
        # Subtract all the geometries for current layer from the fill/ground plane.
        key = (layer_num, data_type)
        if key in self.fill_info:
            self.current_app.modeler.subtract(
                self.fill_info[key]['box_name'],
                self.chip_subtract_dict[layer_num],
                keepOriginals=False)

        # if is_rectangle(qc_shapely):  # Draw as rectangle
        #     self.logger.debug(f"Drawing a rectangle: {name}")
        #     x_min, y_min, x_max, y_max = qc_shapely.bounds
        #     poly_ansys = self.modeler.draw_rect_corner(
        #         *parse_units([
        #             [x_min, y_min,
        #              self.design.get_chip_z(qgeom.chip)],
        #             x_max - x_min,
        #             y_max - y_min,
        #             0,
        #         ]),
        #         **ansys_options,
        #     )
        #     self.modeler.rename_obj(poly_ansys, name)

        # else:
        #     # Draw general closed poly
        #     poly_ansys = self.modeler.draw_polyline(points_3d[:-1],
        #                                             closed=True,
        #                                             **ansys_options)
        #     # rename: handle bug if the name of the cut already exits and is used to make a cut
        #     poly_ansys = poly_ansys.rename(name)

        # qc_fillet = round(qgeom.fillet, 7)
        # if qc_fillet > 0:
        #     qc_fillet = parse_units(qc_fillet)
        #     idxs_to_fillet = good_fillet_idxs(
        #         points,
        #         qc_fillet,
        #         precision=self.design._template_options.PRECISION,
        #         isclosed=True,
        #     )
        #     if idxs_to_fillet:
        #         self.modeler._fillet(qc_fillet, idxs_to_fillet, poly_ansys)

        # # Subtract interior shapes, if any
        # if len(qc_shapely.interiors) > 0:
        #     for i, x in enumerate(qc_shapely.interiors):
        #         interior_points_3d = to_vec3D(parse_units(list(x.coords)),
        #                                       qc_chip_z)
        #         inner_shape = self.modeler.draw_polyline(
        #             interior_points_3d[:-1], closed=True)
        #         self.modeler.subtract(name, [inner_shape])

        # # Input chip info into self.chip_subtract_dict
        # if qgeom.chip not in self.chip_subtract_dict:
        #     self.chip_subtract_dict[qgeom.chip] = set()

        # if qgeom["subtract"]:
        #     self.chip_subtract_dict[qgeom.chip].add(name)

    # Still implementing
    def auto_wirebonds(self, table):
        """
        Adds wirebonds to the Ansys model for path elements where;
        subtract = True and wire_bonds = True.

        Uses render options for determining of the:
        * wb_threshold -- the minimum distance between two vertices of a path for a
        wirebond to be added.
        * wb_offset -- offset distance for wirebond placement (along the direction
        of the cpw)
        * wb_size -- controls the width of the wirebond (wb_size * path['width'])
        """
        # norm_z = np.array([0, 0, 1])

        # wb_threshold = parse_units(self._options["wb_threshold"])
        # wb_offset = parse_units(self._options["wb_offset"])

        # # selecting only the qgeometry which meet criteria
        # wb_table = table.loc[table["hfss_wire_bonds"] == True]
        # wb_table2 = wb_table.loc[wb_table["subtract"] == True]

        # # looping through each qgeometry
        # for _, row in wb_table2.iterrows():
        #     geom = row["geometry"]
        #     width = row["width"]
        #     # looping through the linestring of the path to determine where WBs should be
        #     for index, i_p in enumerate(geom.coords[:-1], start=0):
        #         j_p = np.asarray(geom.coords[:][index + 1])
        #         vert_distance = parse_units(distance.euclidean(i_p, j_p))
        #         if vert_distance > wb_threshold:
        #             # Gets number of wirebonds to fit in section of path
        #             wb_count = int(vert_distance // wb_threshold)
        #             # finds the position vector
        #             wb_pos = (j_p - i_p) / (wb_count + 1)
        #             # gets the norm vector for finding the orthonormal of path
        #             wb_vec = wb_pos / np.linalg.norm(wb_pos)
        #             # finds the orthonormal (for orientation)
        #             wb_perp = np.cross(norm_z, wb_vec)[:2]
        #             # finds the first wirebond to place (rest are in the loop)
        #             wb_pos_step = parse_units(wb_pos + i_p) + (wb_vec *
        #                                                        wb_offset)
        #             # Other input values could be modified, kept to minimal selection for automation
        #             # for the time being. Loops to place N wirebonds based on length of path section.
        #             for wb_i in range(wb_count):
        #                 self.modeler.draw_wirebond(
        #                     pos=wb_pos_step + parse_units(wb_pos * wb_i),
        #                     ori=wb_perp,
        #                     width=parse_units(width * self._options["wb_size"]),
        #                     height=parse_units(width *
        #                                        self._options["wb_size"]),
        #                     z=0,
        #                     wire_diameter="0.015mm",
        #                     NumSides=6,
        #                     name="g_wb",
        #                     material="pec",
        #                     solve_inside=False,
        #                 )
        a = 5

    def get_chip_names(self) -> List[str]:
        """
        Obtain a list of chips on which the selection of components, if valid, resides.

        Returns:
            List[str]: Chips to render.
        """
        if self.case == 2:  # One or more components not in QDesign.
            self.logger.warning('One or more components not found.')
            return []
        chip_names = set()
        if self.case == 1:  # All components rendered.
            comps = self.design.components
            for qcomp in comps:
                if 'chip' not in comps[qcomp].options:
                    self.chip_designation_error()
                    return []
                # elif comps[qcomp].options.chip != 'main':
                #     #self.chip_not_main()
                #     return []
                chip_names.add(comps[qcomp].options.chip)
        else:  # Strict subset rendered.
            icomps = self.design._components
            for qcomp_id in self.qcomp_ids:
                if 'chip' not in icomps[qcomp_id].options:
                    self.chip_designation_error()
                    return []
                # elif icomps[qcomp_id].options.chip != 'main':
                #     #self.chip_not_main()
                #     return []
                chip_names.add(icomps[qcomp_id].options.chip)
        return list(chip_names)

    def get_open_pins_in_layer(self, open_pins: list, layer_num: int) -> list:
        """Reduce the open_pins list and only return the open_pins for layer_nums.

        Args:
            open_pins (list): The list given by user for render_design.
            layer_num (int): The layer number being rendered.

        Returns:
            list: A subset of open_pins for the denoted layer_num.
        """
        layer_open_pins_list = list()

        mask_comp_pin = self.path_and_poly_with_valid_comps[
            'layer'] == layer_num
        path_poly_layer = self.path_and_poly_with_valid_comps[mask_comp_pin]

        for comp_name, pin_name in open_pins:
            id = self.design.components[comp_name].id
            comp_pin_layer_mask = (path_poly_layer['component'] == id)
            valid_df = path_poly_layer[comp_pin_layer_mask]
            if not valid_df.empty:
                layer_open_pins_list.append((comp_name, pin_name))

        return layer_open_pins_list

    def confirm_open_pins_are_valid_names(self, open_pins: list,
                                          pair_used: list) -> bool:
        """Check if all names of components and corresponding pins are
        within design.  Otherwise log an error

        Args:
            open_pins (list): The user input for open_pins.
            pair_used (list): Passed to multiple checks to ensure (component_name, pin)
                            are not accidentally reused.

        Returns:
            bool: True if all components and pins found in design.
                 False if any one component or pin is not found in design.
        """

        for item in open_pins:
            comp_name, pin = item
            if self.qcomp_ids:
                # User wants to render a subset, ensure part of selection.
                if self.design.components[comp_name].id not in self.qcomp_ids:
                    self.design.logger.error(
                        f'You entered the component_name='
                        f'{self.design.components[comp_name].name} and the corresponding'
                        f' name is not selection.'
                        f'The geometries will not be rendered to Ansys.')
                    return False
            if comp_name not in self.design.components.keys():
                self.open_pin_names_not_valid(comp_name, pin)
                return False

            if pin not in self.design.components[comp_name].pins.keys():
                self.open_pin_names_not_valid(comp_name, pin)
                return False
            if item in pair_used:
                self.design.logger.error(
                    f'You entered the component_name={comp_name} and pin={pin} pair twice.  '
                    f'The geometries will not be rendered to Ansys.')
                return False

            pair_used.append(item)
        return True

    def activate_user_project_design(self):
        """If Ansys is not started, then start using the project and name defined by user.
        Otherwise, activate the project and design that user chose. It appears that pyaedt
        can identify the project, if needed, we can pro-actively activate the project.
        """
        try:
            if self.current_app:
                self.current_app.set_active_design(self.design_name)
            else:
                self._initiate_renderer()  # populates self._desktop
                self.populate_project_and_design()

        except Exception as ex:
            self.design.logger.error(
                f'Was not able to activate user\'s project and design name.  The exception is {ex}'
                f'\n Project Name:{self.project_name}, Design Name: {self.design_name}'
            )

    def populate_project_and_design(self):
        """Add project and design based on user input.
        """
        if self.renderer_type in self.renderer_types:
            if self.project_name is None:
                # Can use for future reference.
                #from pyaedt import generate_unique_project_name
                self.project_name = f'{self._get_child_class_name()}_project'
            if self.design_name is None:
                self.design_name = f'{self._get_child_class_name()}_design'
            if self.renderer_type == 'Q3D':
                self.current_app = Q3d(self.project_name, self.design_name)
            elif self.renderer_type == 'HFSS_DM':
                # Default solution type is driven modal.
                self.current_app = Hfss(self.project_name,
                                        self.design_name,
                                        solution_type='Modal')
            elif self.renderer_type == 'HFSS_EM':
                # need to explicitly identify a solution type
                self.current_app = Hfss(self.project_name,
                                        self.design_name,
                                        solution_type='Eigenmode')
        else:
            self.design.logger.error(
                f'Need to implement {self.renderer_type.upper()} in pyaedt_base.'
            )

        self.current_app.set_active_design(self.design_name)
        # Set project_name and design_name based off of the app that was just created
        self.project_name = self.current_app.project_name
        self.design_name = self.current_app.design_name

        self.current_app.modeler.model_units = "mm"
        self.current_app.modeler.oeditor.SetModelUnits(
            ["NAME:Units Parameter", "Units:=", "mm", "Rescale:=", False])

    def clean_user_design(self):
        """Clean the user project and design corresponding to the current_app

        Args: None

        Returns: None
        """
        if self._desktop and self.current_app:
            all_objects = []
            if self.project_name and self.design_name:
                self.activate_user_project_design()
                for group in self.GROUPS:
                    object_list = self.current_app.modeler.get_objects_in_group(
                        group)
                    if object_list:
                        all_objects.extend(object_list)
                self.current_app.modeler.delete(all_objects)

    ######### Warnings and Errors##################################################

    def chip_designation_error(self):
        """
        Warning message that appears when the Ansys renderer fails to locate a component's chip designation.
        Provides instructions for a temporary workaround until the layer stack is finalized.
        """
        self.logger.warning(
            "This component currently lacks a chip designation. Please add chip='main' to the component's default_options dictionary, restart the kernel, and try again."
        )

    def chip_not_main(self):
        """
        Warning message that appears when a component's chip designation is not 'main'.
        As of 05/10/21, all chip designations should be 'main' until the layer stack is finalized.
        Provides instructions for a temporary workaround until the layer stack is finalized.
        """
        self.logger.warning(
            "The chip designation for this component is not 'main'. Please set chip='main' in its default_options dictionary, restart the kernel, and try again."
        )

    def open_pin_names_not_valid(self, comp_name: str, pin_name: str):
        """Stop and give error if the component name or pin name from open_pins
        are not valid within design.

        Args:
            comp_name (str): Name of component which should have an open pin.
            pin_name (str): Name of pin which should be open.
        """
        self.logger.error(
            f'\n The component name or pin name within list'
            f'\nprovided to render_design are not valid.'
            f'\nYou provided: {comp_name} and {pin_name}'
            f'\n Valid components: {self.design.components.keys()}'
            f'\n Valid pins:{self.design.components[comp_name].pins.keys()}'
            f'\n ALL tuples within open_pins will not be addressed.')

    #########PROPERTIES##################################################

    @property
    def desktop(self) -> Desktop:
        """Access to hidden variable for Ansys desktop.

        Returns:
            Desktop: easier reference to pyaedt desktop.
        """

        return self._desktop



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""For GDS export, separate the logic for cheesing."""

import logging
from typing import Union

import gdstk
import numpy as np
import shapely


class Cheesing:
    """Create a cheese cell based on input of no-cheese locations."""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    # pylint: disable=too-few-public-methods

    # To be used by QGDSRenderer only.
    # Number of instance attributes is acceptable for this case.

    def __init__(
        self,
        multi_poly: shapely.geometry.multipolygon.MultiPolygon,
        all_nocheese_gds: list,
        lib: gdstk.Library,
        minx: float,
        miny: float,
        maxx: float,
        maxy: float,
        chip_name: str,
        edge_nocheese: float,
        layer: int,
        is_neg_mask: bool,
        datatype_cheese: int,
        datatype_keepout: int,
        fab: bool,
        logger: logging.Logger,
        max_points: int,
        precision: float,
        cheese_shape: int = 0,
        #  For rectangle
        shape_0_x: float = 0.000050,
        shape_0_y: float = 0.000050,
        #  For Circle
        shape_1_radius: float = 0.000025,
        # delta spacing for holes
        delta_x: float = 0.00010,
        delta_y: float = 0.00010,
    ):
        """Create cheesing polygons based on the no-cheese regions.

        Args:
            multi_poly: No-cheese areas per layer (shapely multipolygon).
            all_nocheese_gds: Same geometry in a gdstk-friendly list.
            lib: Library that will hold all generated cells.
            minx: Chip minimum x location.
            miny: Chip minimum y location.
            maxx: Chip maximum x location.
            maxy: Chip maximum y location.
            chip_name: User-defined chip name.
            edge_nocheese: Buffer around the chip perimeter to leave un-cheesed.
            layer: Layer number for calculating the cheese.
            is_neg_mask: If True, export a negative mask; otherwise positive.
            datatype_cheese: Datatype for cheese polygons.
            datatype_keepout: Datatype for keepout regions.
            fab: If True, output fabrication-ready layers; if False, include
                intermediate/debug layers.
            max_points: Maximum number of points for a polygon in gdstk.
            precision: gdstk precision to use.
            logger: Logger to emit warnings/errors.
            cheese_shape: 0 for rectangle, 1 for circle.
            shape_0_x: Rectangle width (centered at 0,0).
            shape_0_y: Rectangle height (centered at 0,0).
            shape_1_radius: Circle radius.
            delta_x: Spacing between holes in x.
            delta_y: Spacing between holes in y.
        """

        # All the no-cheese locations.
        self.multi_poly = multi_poly
        self.nocheese_gds = all_nocheese_gds
        self.lib = lib

        # chip boundary, layer and datatype, buffer for perimeter
        self.minx = minx
        self.miny = miny
        self.maxx = maxx
        self.maxy = maxy

        self.chip_name = chip_name

        self.edge_nocheese = edge_nocheese
        self.layer = layer
        self.is_neg_mask = is_neg_mask
        self.datatype_cheese = datatype_cheese
        self.datatype_keepout = datatype_keepout
        self.max_points = max_points
        self.precision = precision

        self.fab = fab

        self.logger = logger

        # Expect to mostly cheese a square, but allow for expansion.
        self.cheese_shape = cheese_shape
        self.shape_0_x = shape_0_x
        self.shape_0_y = shape_0_y
        self.shape_1_radius = shape_1_radius

        # Create a shapely the size of chip.
        self.boundary = shapely.geometry.Polygon([(minx, miny), (minx, maxy),
                                                  (maxx, maxy)])

        self.delta_x = delta_x
        self.delta_y = delta_y

        self.cheese_cell = None

        # max dimension of grid is chip size reduced by self.edge_nocheese
        self.grid_minx = self.minx + self.edge_nocheese
        self.grid_miny = self.miny + self.edge_nocheese
        self.grid_maxx = self.maxx - self.edge_nocheese
        self.grid_maxy = self.maxy - self.edge_nocheese

        if self.grid_maxx <= self.grid_minx or self.grid_maxy <= self.grid_miny:
            self.logger.warning(
                "When edge_nocheese is applied to decrease the chip size, of where cheesing ",
                "will happen, the resulting size is not realistic.",
            )

        self.one_hole_cell = None

        self.hole = None

    def apply_cheesing(self) -> gdstk.Library:
        """Prototype, not complete.

        Need to populate self.lib with cheese holes.
        """

        if self._error_checking_hole_delta() == 0:
            # Place hole into self.hole
            self._make_one_hole_at_zero_zero()

            _ = self._hole_to_lib()
            self._cell_with_grid()
        else:
            self.logger.warning("Cheesing not implemented.")

        return self.lib

    def _error_checking_hole_delta(self) -> int:
        """Check ratio of hole size vs hole spacing.

        Returns:
            int: Observation based on hole size and spacing.

            * 0 No issues detected.
            * 1 Delta spacing less than or equal to hole
            * 2 cheese_shape is unknown to Cheesing class.
        """
        observe = -1
        if self.cheese_shape == 0:
            observe = (1 if self.delta_x <= self.shape_0_x or
                       self.delta_y <= self.shape_0_y else 0)
        elif self.cheese_shape == 1:
            diameter = 2 * self.shape_1_radius
            return 1 if self.delta_x <= diameter or self.delta_y <= diameter else 0
        else:
            self.logger.warning(
                f"The cheese_shape={self.cheese_shape} is unknown in Cheesing class."
            )
            return 2

        if observe == 1:
            self.logger.warning(
                "The size of delta spacing is same as or smaller than hole.")

        return observe

    def _make_one_hole_at_zero_zero(self):
        """This method will create just one hole used for cheesing defined by a
        shapely object.

        It will be placed in self.hole.
        """
        if self.cheese_shape == 0:
            width, height = self.shape_0_x, self.shape_0_y
            self.hole = shapely.geometry.box(-width / 2, -height / 2, width / 2,
                                             height / 2)
        elif self.cheese_shape == 1:
            self.hole = shapely.geometry.Point(0, 0).buffer(self.shape_1_radius)
        else:
            self.logger.warning(
                f"The cheese_shape={self.cheese_shape} is unknown in Cheesing class."
            )

    def _hole_to_lib(self) -> gdstk.Polygon:
        """Convert the self.hole to a gds cell and add to self.lib.
        Put the hole on datatype_cheese +2.  This is expected to change when
        we agree to some convention.

        Returns:
            gdstk.Polygon: The gdstk polygon for single hole for cheesing. None is not made.
        """
        a_poly = None
        if isinstance(self.hole, shapely.geometry.Polygon):
            exterior_poly = gdstk.Polygon(
                list(self.hole.exterior.coords),
                layer=self.layer,
                datatype=self.datatype_cheese + 2,
            )

            # If polygons have a holes, need to remove (subtract) it for gdstk.
            all_interiors = list()
            geom = self.hole
            if geom.interiors:
                for inside in geom.interiors:
                    interior_coords = list(inside.coords)
                    all_interiors.append(interior_coords)
                a_poly_set = gdstk.PolygonSet(all_interiors,
                                              layer=self.layer,
                                              datatype=self.datatype_cheese + 2)
                a_poly = gdstk.boolean(
                    exterior_poly,
                    a_poly_set,
                    "not",
                    precision=self.precision,
                    layer=self.layer,
                    datatype=self.datatype_cheese + 2,
                )
            else:
                a_poly = exterior_poly.fracture(max_points=self.max_points,
                                                precision=self.precision)
        else:
            hole_type = type(self.hole)
            self.logger.warning(f"The self.hole was not converted to gdstk; "
                                f"the type '{hole_type}' was not handled.")

        # convert a_poly to cell, then use cell reference to add to all the cheese in chip_rect_gds
        chip_layer_only_top_name = f"TOP_{self.chip_name}_{self.layer}"
        cheese_one_hole_cell_name = f"TOP_{self.chip_name}_{self.layer}_one_hole"
        self.one_hole_cell = self.lib.new_cell(cheese_one_hole_cell_name)
        self.one_hole_cell.add(*a_poly)

        if self.one_hole_cell.bounding_box() is not None:
            next(
                (c for c in self.lib.cells if c.name == chip_layer_only_top_name
                )).add(gdstk.Reference(self.one_hole_cell))
        else:
            self.lib.remove(self.one_hole_cell)

        return a_poly

    def _cell_with_grid(self):
        """Use the hole at self.one_hole_cell to create a grid.

        Then use the no_cheese cell to remove the holes from grid.  The
        difference will be used subtract from the ground layer with
        geometry. The cells are added to the Top_<chip_name>.
        """

        gather_holes_cell = self._get_all_holes()
        diff_holes_cell = self._subtract_keepout_from_hole_grid(
            gather_holes_cell)
        self.lib.remove(gather_holes_cell)

        if self.is_neg_mask:
            # negative mask for given chip and layer
            self._move_to_under_top_chip_layer_name(diff_holes_cell)
            if self.fab:
                self._both_pos_and_neg_mask_fab()

                # Need the diff cell for negative mask.
                # self._remove_cheese_diff_cell()
        else:
            # positive mask for given chip and layer
            if self.fab:
                self._subtract_from_ground_and_move_under_top_chip_layer(
                    diff_holes_cell)
                self._both_pos_and_neg_mask_fab()

                #  This is something special still to do.
                self._remove_cheese_diff_cell()
                self._remove_ground_chip_layer()
            else:
                self._subtract_from_ground_and_move_under_top_chip_layer(
                    diff_holes_cell)

    def _both_pos_and_neg_mask_fab(self):
        """For both positive and negative mask need to have this cell removed when
        user has fabricate.fab=True.
        """
        self._remove_cell_one_hole()

    def _remove_cell_one_hole(self):
        """Remove cell with just one hole."""
        cheese_one_hole_cell_name = f"TOP_{self.chip_name}_{self.layer}_one_hole"
        cheese_one_hole_cell = next(
            (c for c in self.lib.cells if c.name == cheese_one_hole_cell_name),
            None)
        if cheese_one_hole_cell:
            self.lib.remove(cheese_one_hole_cell)

    def _subtract_from_ground_and_move_under_top_chip_layer(
            self, diff_holes_cell: gdstk.Cell):
        """Get the existing chip_only_top_name cell, then add the holes to it.
        Also, add ground_cheesed_cell under chip_only_top_name

        Args:
            diff_holes_cell (gdstk.Cell): New cell with cheesed ground
        """

        chip_only_top_layer_name = f"TOP_{self.chip_name}_{self.layer}"
        chip_only_top_layer_cell = next(
            (c for c in self.lib.cells if c.name == chip_only_top_layer_name),
            None)
        if chip_only_top_layer_cell:
            if diff_holes_cell.bounding_box() is not None:
                chip_only_top_layer_cell.add(gdstk.Reference(diff_holes_cell))
                ground_cheese_cell = self._subtract_holes_from_ground(
                    diff_holes_cell)

                # Move to under Top_main_layer (Top_chipname_#)
                self._move_to_under_top_chip_layer_name(ground_cheese_cell)
            else:
                self.lib.remove(diff_holes_cell)

    def _subtract_keepout_from_hole_grid(
            self, gather_holes_cell: gdstk.Cell) -> gdstk.Cell:
        """Given a cell with all the holes, subtract the keepout region.
        Then return a new cell with the result.

        Args:
            gather_holes_cell (gdstk.Cell): Holds a grid of all
                                                the holes for cheesing.

        Returns:
            gdstk.Cell: Newly created cell that holds the difference
                                        of holes minus the keep=out region.
        """

        # subtact the keepout, note, Based on user options,
        # the keepout (no_cheese) cell may not be in self.lib.
        temp_keepout_chip_layer_cell = f"temp_keepout_{self.chip_name}_{self.layer}"
        temp_keepout_cell = self.lib.new_cell(temp_keepout_chip_layer_cell)
        temp_keepout_cell.add(*self.nocheese_gds)
        diff_holes = gdstk.boolean(
            gather_holes_cell.get_polygons(),
            temp_keepout_cell.get_polygons(),
            "not",
            precision=self.precision,
            layer=self.layer,
            datatype=self.datatype_cheese + 1,
        )
        diff_holes_cell_name = f"TOP_{self.chip_name}_{self.layer}_Cheese_diff"
        diff_holes_cell = self.lib.new_cell(diff_holes_cell_name)
        diff_holes_cell.add(*diff_holes)

        self.lib.remove(temp_keepout_cell)
        return diff_holes_cell

    def _get_all_holes(self) -> gdstk.Cell:
        """Return a cell with a grid of holes. The keepout has not been
        applied yet.

        Returns:
            gdstk.Cell: Cell containing all the holes.
        """
        gather_holes_cell_name = f"Gather_holes_{self.chip_name}_{self.layer}"
        gather_holes_cell = self.lib.new_cell(gather_holes_cell_name)

        x_holes = np.arange(self.grid_minx,
                            self.grid_maxx,
                            self.delta_x,
                            dtype=float).tolist()
        y_holes = np.arange(self.grid_miny,
                            self.grid_maxy,
                            self.delta_y,
                            dtype=float).tolist()

        if self.one_hole_cell is not None:
            for x_loc in x_holes:
                for y_loc in y_holes:
                    gather_holes_cell.add(
                        gdstk.Reference(self.one_hole_cell,
                                        origin=(x_loc, y_loc)))

        return gather_holes_cell

    def _subtract_holes_from_ground(
            self, diff_holes_cell: gdstk.Cell) -> Union[gdstk.Cell, None]:
        """Get reference to ground cell and then subtract the holes from
        ground. Place the difference into a new cell, which will eventually
        be added under Top.

        Args:
            diff_holes_cell ([type]): Cell which contains all the holes.

        Returns:
            Union[gdstk.Cell, None]: If worked, the new cell with
            cheesed ground, otherwise, None.
        """

        # Still need to 'not' with Top_main_1 (ground)
        top_chip_layer_name = f"TOP_{self.chip_name}_{self.layer}"
        ground_cell_name = f"ground_{self.chip_name}_{self.layer}"
        if next((c for c in self.lib.cells if c.name == top_chip_layer_name),
                None):
            ground_cell = next(
                (c for c in self.lib.cells if c.name == ground_cell_name))
            # Need to keep the depth at 0, otherwise all the
            # cell references (junctions) will be added for boolean.
            ground_cheese = gdstk.boolean(
                ground_cell.get_polygons(depth=0),
                diff_holes_cell.get_polygons(),
                "not",
                precision=self.precision,
                layer=self.layer,
                datatype=self.datatype_cheese,
            )
            ground_cheese_cell_name = (
                f"TOP_{self.chip_name}_{self.layer}_Cheese_{self.datatype_cheese}"
            )
            ground_cheese_cell = self.lib.new_cell(ground_cheese_cell_name)
            return ground_cheese_cell.add(*ground_cheese)

        self.logger.warning(
            f"The cell:{top_chip_layer_name} was not found in self.lib. "
            f"Cheesing not implemented.")
        return None

    def _move_to_under_top_chip_layer_name(self, a_cell: gdstk.Cell):
        """Move the cell to under TOP_<chip name>_<layer number>.

        Args:
            a_cell (gdstk.Cell): A GDSTK cell.
        """
        chip_only_top_chip_layer_name = f"TOP_{self.chip_name}_{self.layer}"
        chip_only_top_chip_layer_cell = next(
            (c for c in self.lib.cells
             if c.name == chip_only_top_chip_layer_name), None)
        if chip_only_top_chip_layer_cell:
            if a_cell.bounding_box() is not None:
                chip_only_top_chip_layer_cell.add(gdstk.Reference(a_cell))
            else:
                self.lib.remove(a_cell)

    def _remove_cheese_diff_cell(self):
        """For a lib, chip and layer, remove the Cheese_diff cell."""
        cell_name = f"TOP_{self.chip_name}_{self.layer}_Cheese_diff"
        cell = next((c for c in self.lib.cells if c.name == cell_name), None)
        if cell:
            self.lib.remove(cell)

    def _remove_ground_chip_layer(self):
        """[For a lib, chip and layer, remove the ground cell
        which is created for positive mask.
        """
        cell_name = f"ground_{self.chip_name}_{self.layer}"
        cell = next((c for c in self.lib.cells if c.name == cell_name), None)
        if cell:
            self.lib.remove(cell)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from qiskit_metal import draw
from qiskit_metal.toolbox_python.attr_dict import Dict
from qiskit_metal.qlibrary.core import QComponent


class Cap3Interdigital(QComponent):
    """Create a three finger planar capacitor with a ground pocket cuttout.
    The width of the fingers is determined by the trace width.

    Inherits QComponent class.

    Capacitor Metal Geometry and Ground Cutout Pocket:
        * finger length  - length of each finger
        * pocket_buffer_width_x - sets size of pocket in +-x direction, added to cap size
        * pocket_buffer_width_y - sets size of pocket in +-y direction, added to cap size
          this also determines the lead in line lengths
          pocket is a negative shape that is cut out of the ground plane

    Pins:
        There are two pins on the capacitor at either end
        The pins attach directly to the built in lead length and only needs a width defined
        * trace_width - center trace width of the trace lead line and cap fingers

    Sketch:
        Below is a sketch of the capacitor
        ::

                 |
            -----------
                 |
            |    |    |
            |    |    |
            |    |    |
            |    |    |
            |         |
            -----------
                 |

    .. image::
        Cap3Interdigital.png

    .. meta::
        :description: Cap 3 Interdigital

    Default Options:
        * trace_width: '10um'
        * finger_length: '65um'
        * pocket_buffer_width_x: '10um'
        * pocket_buffer_width_y: '30um'
    """

    #  Define structure functions

    default_options = Dict(trace_width='10um',
                           finger_length='65um',
                           pocket_buffer_width_x='10um',
                           pocket_buffer_width_y='30um')
    """Default drawing options"""

    TOOLTIP = """Create a three finger planar capacitor with a ground pocket cuttout."""

    def make(self):
        """This is executed by the user to generate the qgeometry for the
        component."""
        p = self.p
        #########################################################

        # Make the shapely polygons for the main cap structure
        pad = draw.rectangle(p.trace_width * 5, p.trace_width)
        pad_top = draw.translate(pad, 0,
                                 +(p.trace_width * 2 + p.finger_length) / 2)
        pad_bot = draw.translate(pad, 0,
                                 -(p.trace_width * 2 + p.finger_length) / 2)
        finger = draw.rectangle(p.trace_width, p.finger_length)
        cent_finger = draw.translate(finger, 0, +(p.trace_width) / 2)
        left_finger = draw.translate(finger, -(p.trace_width * 2),
                                     -(p.trace_width) / 2)
        right_finger = draw.translate(finger, +(p.trace_width * 2),
                                      -(p.trace_width) / 2)

        # Make the shapely polygons for the leads in the pocket (length=pocket_buffer_width_y)
        trace_temp_1 = draw.rectangle(p.trace_width, p.pocket_buffer_width_y)
        trace_top = draw.translate(
            trace_temp_1, 0,
            +(p.finger_length + p.trace_width * 3 + p.pocket_buffer_width_y * 2)
            / 2 - p.pocket_buffer_width_y / 2)
        trace_bot = draw.translate(
            trace_temp_1, 0,
            -(p.finger_length + p.trace_width * 3 + p.pocket_buffer_width_y * 2)
            / 2 + p.pocket_buffer_width_y / 2)

        # Make the shapely polygons for pocket ground plane cuttout
        pocket = draw.rectangle(
            p.trace_width * 5 + p.pocket_buffer_width_x * 2,
            p.finger_length + p.trace_width * 3 + p.pocket_buffer_width_y * 2)

        # These variables are used to graphically locate the pin locations
        top_pin_line = draw.LineString([
            (-p.trace_width / 2,
             (p.finger_length + p.trace_width * 3 + p.pocket_buffer_width_y * 2)
             / 2),
            (+p.trace_width / 2,
             (p.finger_length + p.trace_width * 3 + p.pocket_buffer_width_y * 2)
             / 2)
        ])
        bot_pin_line = draw.LineString([(
            -p.trace_width / 2,
            -(p.finger_length + p.trace_width * 3 + p.pocket_buffer_width_y * 2)
            / 2),
                                        (+p.trace_width / 2,
                                         -(p.finger_length + p.trace_width * 3 +
                                           p.pocket_buffer_width_y * 2) / 2)])

        # Create polygon object list
        polys1 = [
            top_pin_line, bot_pin_line, pad_top, pad_bot, cent_finger,
            left_finger, right_finger, pocket, trace_top, trace_bot
        ]

        # Rotates and translates all the objects as requested. Uses package functions
        # in 'draw_utility' for easy rotation/translation
        polys1 = draw.rotate(polys1, p.orientation, origin=(0, 0))
        polys1 = draw.translate(polys1, xoff=p.pos_x, yoff=p.pos_y)
        [
            top_pin_line, bot_pin_line, pad_top, pad_bot, cent_finger,
            left_finger, right_finger, pocket, trace_top, trace_bot
        ] = polys1

        # Adds the object to the qgeometry table
        self.add_qgeometry('poly',
                           dict(pad_top=pad_top,
                                pad_bot=pad_bot,
                                cent_finger=cent_finger,
                                left_finger=left_finger,
                                right_finger=right_finger,
                                trace_top=trace_top,
                                trace_bot=trace_bot),
                           layer=p.layer)

        #subtracts out ground plane on the layer its on
        self.add_qgeometry('poly',
                           dict(pocket=pocket),
                           subtract=True,
                           layer=p.layer)

        # Generates its own pins
        self.add_pin('a', top_pin_line.coords, p.trace_width)
        self.add_pin('b', bot_pin_line.coords[::-1], p.trace_width)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Parsed dynamic attributes."""
import pprint
from typing import List
from typing import TYPE_CHECKING
#from ...toolbox_python.utility_functions import log_error_easy
if TYPE_CHECKING:
    from .base import QComponent


def is_ipython_magic(check_attribute: str) -> bool:
    """Ignore the following checks by jupyter."""
    return check_attribute in {
        '_ipython_canary_method_should_not_exist_',
        '_ipython_display_',
        '_repr_mimebundle_',
        '_repr_html_',
        '_repr_markdown_',
        '_repr_svg_',
        '_repr_png_',
        '_repr_pdf_',
        '_repr_jpeg_',
        '_repr_latex_',
        '_repr_json_',
        '_repr_javascript_',
        '_rapped',
        '__wrapped__',
        '__call__',
    }


class ParsedDynamicAttributes_Component():
    """Provides a parsing view of the component options.

    When accessed, returns parse versions of the user options.
    Works with nested options too.

    Example:
        .. code-block:: python

            component.options = {'x':'1nm'}
            print(component.p.x)

        >> `float(1e7)`
    """
    """
    Special method names:
        See Python Data Module for Objects:
            https://docs.python.org/3/reference/datamodel.html
            https://docs.python.org/3/library/stdtypes.html#object.__dict__
            https://python-reference.readthedocs.io/en/latest/docs/dunderattr/index.html
            https://rszalski.github.io/magicmethods/

        __getattribute__:        Called unconditionally to implement attribute accesses for instances of the class.
        __getattr__:        Called when an attribute lookup has not found the attribute in the usual places.
        __setattr__:        Called when an attribute assignment is attempted.
        __delattr__:        Called when an attribute deletion is attempted.
    """

    def __init__(self, component: 'QComponent', key_list: List[str] = None):
        """
        Args:
            component (QComponent): Component to get options from.
            key_list (List[str]): List of keys.  Defaults to None.
        """
        #print(f'*** Created with {key_list}')
        # These names must have __xx__ or else they will go to getattr instead of getattribute

        self.__component__ = component
        # self.__d__ = component.options
        self.__keylist__ = key_list or []  # lis tot get current value

        self.__parse__ = component.design.parse_value  # function

    def __dir__(self):
        # For autocompletion
        return list(self.__getdict__().keys())

    # def _repr_html_(self):
    #    return self.__d__._repr_html_()

    def __str__(self):
        return self.__getdict__().__str__()

    def __len__(self):
        """Return the length."""
        return len(self.__getdict__())

    def __getdict__(self) -> dict:
        return get_nested_dict_item(self.__component__.options,
                                    self.__keylist__)

    def __iter__(self):
        return self.__getdict__().__iter__()

    def items(self):
        """Produces tuples consisting of keys and respective parsed values for
        iterating over a dictionary."""
        for key in self:  # calls __iter__
            yield (key, self.__getitem__(key))

    def __getattr__(self, name: str):
        """Delegating Attribute Access After regular attribute access, try
        looking up the name This allows simpler access to columns for
        interactive use."""
        #### Note: obj.x will always call obj.__getattribute__('x') prior to
        # calling obj.__getattr__('x').
        #
        #### Former issues:
        # These two only come up when trying to use in ipython /jupyter environemtn
        # and acessing a sub dictionary in the past
        # Exmaple:
        #   xx = ParsedDynamicAttributes(self, ['c'])
        #   xx
        # if name.startswith('_repr') or name.startswith('_ipython'):
        #    return

        #print(f'__getattr__ NAME = {name};  dict=',self.__getdict__())

        return self.__getitem__(name)

    def __getitem__(self, name: str):
        """Get the item associated with the given name.

        Args:
            name (str): Name of the item

        Return:
            str, float, list, tuple, or ast eval: Parse value

        Raises:
            AttributeError: The given name is a magic method not in the dictionary
        """

        dic = self.__getdict__()
        if name not in dic:
            if not is_ipython_magic(name):
                # log_error_easy(self.__component__.logger, post_text=
                xx = self.__keylist__
                xx = ".".join(xx) + "." + str(name) if len(xx) > 0 else name
                self.__component__.logger.error(
                    '\nWarning: User tried to access a variable in the parse options'
                    f' that is not there!\n Component name = `{self.__component__.name}`\n'
                    f' Option name    = `{xx}`')
                return None
            else:
                # IPython checking methods
                # https://github.com/jupyter/notebook/issues/2014
                raise AttributeError(name)

        else:
            val = dic.get(name)
            #print(f'val = {val}')
            if isinstance(val, dict):
                #print(f'  -> Going to create a new {self.__keylist__ + [name]}')
                return ParsedDynamicAttributes_Component(
                    self.__component__, key_list=self.__keylist__ + [name])
            else:
                return self.__parse__(val)

    ####### SERIALIZATION

    def __getstate__(self):
        #  "I'm being pickled"
        return self.__dict__

    def __setstate__(self, d):
        #  f"I'm being unpickled with these values: {d}"
        self.__dict__ = d

    # def __repr__(self):
    #     return self.__getdict__().__repr__()

    def __repr__(self):
        """For viewing.

        Just return the whole parse dictionary.
        """
        b = '\033[95m\033[1m'
        e = '\033[0m'

        c = self.__component__
        parsed = c.parse_value(c.options)
        parsed_text = pprint.pformat(parsed, indent=1)
        text = f"""{b}Current parsed *view* of options for {c.name}:{e}
 (Units are parsed to floats in default design units.)
 {parsed_text}"""
        return text


# TESTING code:
"""
from qiskit_metal.qlibrary.core.base import QComponent

class Test(QComponent):
    def make(self):
        pass
DEFAULT_OPTIONS['Test'] = {     #instead of using DEFAULT_OPTIONS, use the dicts within QDesign class.
    'a' : '1mm',
    'b' : '1um',
    'c' : {
        'd' : '15um',
        'e' : '10um',
        }
    }
self = Test(design, 'test')
self.p = ParsedDynamicAttributes(self)
self.p
self.p.a
self.p.a.c
"""


def get_nested_dict_item(dic: dict, key_list: list, level=0):
    """Get the nested dictionary item.

    Args:
        dic (dict): Dictionary of nested items
        key_list (list): List of keys
        level (int): Internal for recussion.  Defaults to 0.

    Returns:
        object: Item in the dic corresponding to the given key_list

    Example use:

        .. code-block: python

            myDict = Dict(
                aa=Dict(
                    x1={
                        'dda': '34fF'
                    },
                    y1='Y',
                    z='10um'
                ),
                bb=Dict(
                    x2=5,
                    y2='YYYsdg',
                    z='100um'
                ),
                cc='100nm'
            )
            key_list = ['cc']
            print(get_nested_dict_item(myDict, key_list))

            key_list = ['aa', 'x1', 'dda']
            print(get_nested_dict_item(myDict, key_list))

        Results in
            >> 100nm
            >> 34fF
    """

    if not key_list:  # get the root
        return dic

    if level < len(key_list) - 1:
        return get_nested_dict_item(dic[key_list[level]], key_list, level + 1)
    else:
        return dic[key_list[level]]


# Example use
"""
myDict = Dict(
    aa=Dict(
        x1={
            'dda': '34fF'
        },
        y1='Y',
        z='10um'
    ),
    bb=Dict(
        x2=5,
        y2='YYYsdg',
        z='100um'
    ),
    cc='100nm'
)
key_list = ['cc']
print(get_nested_dict_item(myDict, key_list))

key_list = ['aa', 'x1', 'dda']
print(get_nested_dict_item(myDict, key_list))

print(get_nested_dict_item(myDict, ['aa']))
"""



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=protected-access
# pylint: disable-msg=relative-beyond-top-level
# pylint: disable-msg=broad-except
"""Saving and load metal data."""

import pickle
#from ..designs.base
from ..toolbox_python.utility_functions import log_error_easy

__all__ = ['save_metal', 'load_metal_design']


def save_metal(filename: str, design):
    """Save metal0.

    Args:
        filename (str): File path
        design (QDesign): Design obejct Metal_Design_Base

    Returns:
        bool: True is sucessful, False otherwise

    Raises:
        Exception: Save error
    """
    result = False

    # Do not pickle the following -- THis is issue with picking
    self = design  # cludge for lazy tying
    logger = self.logger
    self.logger = None

    # Pickle
    # TODO: Right now just does pickle. Need to serialize object into JSON
    # or similar. Then recreate components.
    try:
        pickle.dump(self, open(filename, "wb"))

        result = True
    except Exception as e:
        # handle errors here? such as PermissionError
        text = f'ERROR WHILE SAVING: {e}'
        log_error_easy(logger, post_text=text)
        # print(text)
        # logger.error(text)
        result = False

    # restore -- also need to do in the load function
    self.logger = logger
    return result


# pylint: disable-msg=import-outside-toplevel
def load_metal_design(filename: str):
    """Load metal design.

    Args:
        filename (str): File path

    Returns:
        picked QDesign: The pickled design object and updates if asked the param dicts for defaults
    """
    design = pickle.load(open(filename, "rb"))
    design.save_path = str(
        filename)  # Set the place from where we loaded the design

    # Restore
    from .. import logger
    design.logger = logger  #TODO: fix from save pikcle

    return design



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Tree view for Param Entry Window
"""

from PySide6 import QtGui, QtWidgets
from PySide6.QtCore import QModelIndex
from PySide6.QtWidgets import QTreeView


class TreeViewParamEntry(QTreeView):
    """Handles editing a QComponent

    This class extends the `QTreeView` and `QWidget_PlaceholderText` classes
    """

    def __init__(self, parent: QtWidgets.QWidget):
        """
        Args:
            parent (QtWidgets.QWidget): the widget
        """
        QTreeView.__init__(self, parent)

        # not sure whu the ui isn't updating these here.
        self.expanded.connect(self.resize_on_expand)

    def style_me(self):
        """Set style"""
        self.setStyleSheet("""
QTreeView::branch {  border-image: url(none.png); }
        """)

    def mousePressEvent(self, event: QtGui.QMouseEvent):
        """ Overrides inherited mousePressEvent to allow user to clear any selections
        by clicking off the displayed tree. Then calls the inherited mousePressEvent"""
        myindex = self.indexAt(event.pos())
        if myindex.row() == -1:
            self.clearSelection()
            self.setCurrentIndex(QModelIndex())
        super().mousePressEvent(event)

    def resize_on_expand(self):
        """Resize when exposed"""
        self.resizeColumnToContents(0)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=unnecessary-pass
# pylint: disable-msg=too-many-public-methods
# pylint: disable-msg=protected-access
# pylint: disable-msg=import-error
"""Qiskit Metal unit tests components functionality."""

import unittest
import numpy as np

from qiskit_metal.qlibrary.core import _parsed_dynamic_attrs
from qiskit_metal import Dict
from qiskit_metal import draw
from qiskit_metal.qlibrary._template import MyQComponent
from qiskit_metal.qlibrary.core import QComponent
from qiskit_metal.qlibrary.core import QRoute
from qiskit_metal.qlibrary.core import BaseQubit
from qiskit_metal.qlibrary.lumped.cap_n_interdigital import CapNInterdigital
from qiskit_metal.qlibrary.couplers.coupled_line_tee import CoupledLineTee
from qiskit_metal.qlibrary.couplers.cap_n_interdigital_tee import CapNInterdigitalTee
from qiskit_metal.qlibrary.couplers.line_tee import LineTee
from qiskit_metal.qlibrary.terminations import open_to_ground
from qiskit_metal.qlibrary.terminations import short_to_ground
from qiskit_metal.qlibrary.tlines import anchored_path
from qiskit_metal.qlibrary.tlines.anchored_path import RouteAnchors
from qiskit_metal.qlibrary.tlines.framed_path import RouteFramed
from qiskit_metal.qlibrary.tlines.meandered import RouteMeander
from qiskit_metal.qlibrary.tlines import straight_path
from qiskit_metal import designs
from qiskit_metal.qlibrary.qubits import star_qubit
from qiskit_metal.qlibrary.qubits.JJ_Dolan import jj_dolan
from qiskit_metal.qlibrary.qubits.JJ_Manhattan import jj_manhattan
from qiskit_metal.qlibrary.qubits import transmon_pocket_cl
from qiskit_metal.qlibrary.qubits import transmon_pocket
from qiskit_metal.qlibrary.qubits import transmon_cross
from qiskit_metal.qlibrary.qubits import transmon_cross_fl
from qiskit_metal.qlibrary.qubits import transmon_pocket_6
from qiskit_metal.qlibrary.qubits.transmon_pocket_teeth import TransmonPocketTeeth
from qiskit_metal.qlibrary.qubits.SQUID_loop import SQUID_LOOP
from qiskit_metal.qlibrary.couplers import tunable_coupler_01
from qiskit_metal.tests.assertions import AssertionsMixin

#pylint: disable-msg=line-too-long
from qiskit_metal.qlibrary.lumped.resonator_coil_rect import ResonatorCoilRect


class TestComponentFunctionality(unittest.TestCase, AssertionsMixin):
    """Unit test class."""

    def setUp(self):
        """Setup unit test."""
        pass

    def tearDown(self):
        """Tie any loose ends."""
        pass

    def test_qlibrary_get_nested_dict_item(self):
        """Test the functionality of get_nested_dict_item in
        _parsed_dynamic_attrs.py."""
        # Setup expected test results
        my_dict = Dict(aa=Dict(x1={'dda': '34fF'}, y1='Y', z='10um'),
                       bb=Dict(x2=5, y2='YYYsdg', z='100um'),
                       cc='100nm')

        # Generate actual result data
        _test_a = _parsed_dynamic_attrs.get_nested_dict_item(my_dict, ['cc'])
        _test_b = _parsed_dynamic_attrs.get_nested_dict_item(
            my_dict, ['aa', 'x1', 'dda'])
        _test_c = _parsed_dynamic_attrs.get_nested_dict_item(my_dict, ['aa'])
        _test_d = _parsed_dynamic_attrs.get_nested_dict_item(my_dict, ['empty'])

        # Test all elements of the result data against expected data
        self.assertEqual('100nm', _test_a)
        self.assertEqual('34fF', _test_b)
        self.assertEqual({
            'x1': {
                'dda': '34fF'
            },
            'y1': 'Y',
            'z': '10um'
        }, _test_c)
        self.assertEqual({}, _test_d)

    def test_qlibrary_get_and_set_qcomponent_name(self):
        """Test the getting and setting of a QComponent name."""
        design = designs.DesignPlanar()
        my_qcomponent_local = None

        my_qcomponent_local = QComponent(design, "my_name", make=False)
        self.assertEqual(my_qcomponent_local.name, "my_name")

        my_qcomponent_local.name = "another-name"
        self.assertEqual(my_qcomponent_local.name, "another-name")

    #pylint: disable-msg=too-many-statements
    def test_qlibrary_qcomponent_add_pins(self):
        """Test pin addition."""
        design = designs.DesignPlanar()
        my_q_component = QComponent(design, "my_name-add-pins", make=False)

        spiral_list_0 = self.generate_spiral_list(0, 0)
        spiral_list_1 = self.generate_spiral_list(1, 2)

        my_q_component.add_pin('pin1-name',
                               np.array(spiral_list_0.coords)[-2:], 0.1)
        my_q_component.add_pin('pin2-name',
                               np.array(spiral_list_1.coords)[-2:], 0.1)

        my_pins = my_q_component.pins

        self.assertEqual(len(my_pins), 2)
        self.assertEqual(len(my_pins['pin1-name']), 10)
        self.assertEqual(len(my_pins['pin1-name']['points']), 2)
        self.assertEqual(len(my_pins['pin1-name']['points'][0]), 2)
        self.assertEqual(len(my_pins['pin1-name']['points'][1]), 2)
        self.assertEqual(my_pins['pin1-name']['points'][0][0], -35.)
        self.assertEqual(my_pins['pin1-name']['points'][0][1], 30.)
        self.assertEqual(my_pins['pin1-name']['points'][1][0], -35.)
        self.assertEqual(my_pins['pin1-name']['points'][1][1], -35.)
        self.assertEqual(len(my_pins['pin1-name']['middle']), 2)
        self.assertEqual(my_pins['pin1-name']['middle'][0], -35.)
        self.assertEqual(my_pins['pin1-name']['middle'][1], -2.5)
        self.assertEqual(len(my_pins['pin1-name']['normal']), 2)
        self.assertEqual(my_pins['pin1-name']['normal'][0], 1.)
        self.assertEqual(my_pins['pin1-name']['normal'][1], -0.)
        self.assertEqual(len(my_pins['pin1-name']['tangent']), 2)
        self.assertEqual(my_pins['pin1-name']['tangent'][0], 0.)
        self.assertEqual(my_pins['pin1-name']['tangent'][1], -1.)
        self.assertEqual(my_pins['pin1-name']['width'], 65.0)
        self.assertEqual(my_pins['pin1-name']['gap'], 0.06)
        self.assertEqual(my_pins['pin1-name']['chip'], 'main')
        self.assertEqual(my_pins['pin1-name']['parent_name'], 1)
        self.assertEqual(my_pins['pin1-name']['net_id'], 0)
        self.assertEqual(my_pins['pin1-name']['length'], 0)

        self.assertEqual(len(my_pins['pin2-name']), 10)
        self.assertEqual(len(my_pins['pin2-name']['points']), 2)
        self.assertEqual(len(my_pins['pin2-name']['points'][0]), 2)
        self.assertEqual(len(my_pins['pin2-name']['points'][1]), 2)
        self.assertEqual(my_pins['pin2-name']['points'][0][0], -34.)
        self.assertEqual(my_pins['pin2-name']['points'][0][1], 32.)
        self.assertEqual(my_pins['pin2-name']['points'][1][0], -34.)
        self.assertEqual(my_pins['pin2-name']['points'][1][1], -33.)
        self.assertEqual(len(my_pins['pin2-name']['middle']), 2)
        self.assertEqual(my_pins['pin2-name']['middle'][0], -34.)
        self.assertEqual(my_pins['pin2-name']['middle'][1], -0.5)
        self.assertEqual(len(my_pins['pin2-name']['normal']), 2)
        self.assertEqual(my_pins['pin2-name']['normal'][0], 1.)
        self.assertEqual(my_pins['pin2-name']['normal'][1], -0.)
        self.assertEqual(len(my_pins['pin2-name']['tangent']), 2)
        self.assertEqual(my_pins['pin2-name']['tangent'][0], 0.)
        self.assertEqual(my_pins['pin2-name']['tangent'][1], -1.)
        self.assertEqual(my_pins['pin2-name']['width'], 65.0)
        self.assertEqual(my_pins['pin2-name']['gap'], 0.06)
        self.assertEqual(my_pins['pin2-name']['chip'], 'main')
        self.assertEqual(my_pins['pin2-name']['parent_name'], 1)
        self.assertEqual(my_pins['pin2-name']['net_id'], 0)
        self.assertEqual(my_pins['pin2-name']['length'], 0)
        my_q_component.delete()

    def test_qlibrary_template_component_metadata(self):
        """Test component_metadata in _template.py."""
        component = MyQComponent
        self.assertEqual(component.component_metadata,
                         {'short_name': 'component'})

    def test_qlibrary_base_component_metadata(self):
        """Test component_metadata in base/base.py."""
        component = QComponent
        self.assertEqual(component.component_metadata, {})

    def test_qlibrary_base_get_template_options(self):
        """Test get_template_options in base.py."""
        design = designs.DesignPlanar()

        expected = Dict(pos_x='0.0um',
                        pos_y='0.0um',
                        orientation='0.0',
                        chip='main',
                        layer='1')
        self.assertEqual(QComponent.get_template_options(design), expected)

        expected.update(connection_pads={}, _default_connection_pads={})
        self.assertEqual(BaseQubit.get_template_options(design), expected)

    def test_qlibrary_qubit_component_metadata(self):
        """Test component_metadata in base/qubit.py."""
        component = BaseQubit
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 2)
        self.assertEqual(metadata['short_name'], 'Q')
        self.assertEqual(metadata['_qgeometry_table_poly'], 'True')

    def test_qlibrary_cpw_hanger_t_component_metadata(self):
        """Test component_metadata in component/coupled_line_tee.py."""
        component = CoupledLineTee

        metadata = component.component_metadata
        self.assertEqual(len(metadata), 2)
        self.assertEqual(metadata['short_name'], 'cpw')
        self.assertEqual(metadata['_qgeometry_table_path'], 'True')

    def test_qlibrary_cpw_finger_component_metadata(self):
        """Test component_metadata in component/cap_n_interdigital.py."""
        component = CapNInterdigital

        metadata = component.component_metadata

        self.assertEqual(len(metadata), 3)
        self.assertEqual(metadata['short_name'], 'cpw')
        self.assertEqual(metadata['_qgeometry_table_path'], 'True')
        self.assertEqual(metadata['_qgeometry_table_poly'], 'True')

    def test_qlibrary_cpw_t_finger_component_metadata(self):
        """Test component_metadata in component/cap_n_interdigital_tee.py."""
        component = CapNInterdigitalTee

        metadata = component.component_metadata

        self.assertEqual(len(metadata), 3)
        self.assertEqual(metadata['short_name'], 'cpw')
        self.assertEqual(metadata['_qgeometry_table_path'], 'True')
        self.assertEqual(metadata['_qgeometry_table_poly'], 'True')

    def test_qlibrary_cpw_t_component_metadata(self):
        """Test component_metadata in component/line_tee.py."""
        component = LineTee

        metadata = component.component_metadata

        self.assertEqual(len(metadata), 2)
        self.assertEqual(metadata['short_name'], 'cpw')
        self.assertEqual(metadata['_qgeometry_table_path'], 'True')

    def test_qlibrary_open_to_ground_component_metadata(self):
        """Test component_metadata in component/open_to_ground.py."""
        component = open_to_ground.OpenToGround

        metadata = component.component_metadata
        self.assertEqual(len(metadata), 2)
        self.assertEqual(metadata['short_name'], 'term')
        self.assertEqual(metadata['_qgeometry_table_poly'], 'True')

    def test_qlibrary_short_to_ground_component_metadata(self):
        """Test component_metadata in component/short_to_ground.py."""
        component = short_to_ground.ShortToGround
        self.assertEqual(component.component_metadata, {'short_name': 'term'})

    def test_qlibrary_route_anchors_component_metadata(self):
        """Test component_metadata in RouteAnchors."""
        component = RouteAnchors
        self.assertEqual(component.component_metadata, {'short_name': 'cpw'})

    def test_qlibrary_framed_path_component_metadata(self):
        """Test component_metadata in tlines/framed_path.py."""
        component = RouteFramed
        self.assertEqual(component.component_metadata, {'short_name': 'cpw'})

    def test_qlibrary_straight_path_component_metadata(self):
        """Test component_metadata in tlines/straight_path.py."""
        component = straight_path.RouteStraight
        self.assertEqual(component.component_metadata, {'short_name': 'cpw'})

    def test_qlibrary_meander_path_component_metadata(self):
        """Test component_metadata in tlines/meandered.py."""
        component = RouteMeander
        self.assertEqual(component.component_metadata, {'short_name': 'cpw'})

    def test_qlibrary_qroute_base_component_metadata(self):
        """Test component_metadata in tlines/qroute_base.py."""
        component = QRoute
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 2)
        self.assertEqual(metadata['short_name'], 'route')
        self.assertEqual(metadata['_qgeometry_table_path'], 'True')

    def test_qlibrary_resonator_rectangle_spiral_component_metadata(self):
        """Test component_metadata in
        tlines/resonator_coil_rect.py."""
        component = ResonatorCoilRect
        self.assertEqual(component.component_metadata, {'short_name': 'res'})

    def test_qlibrary_transmon_pocket_cl_component_metadata(self):
        """Test component_metadata in qubits/transmon_pocket_cl.py."""
        component = transmon_pocket_cl.TransmonPocketCL
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 2)
        self.assertEqual(metadata['short_name'], 'Q')
        self.assertEqual(metadata['_qgeometry_table_poly'], 'True')

    def test_qlibrary_transmon_pocket_component_metadata(self):
        """Test component_metadata in qubits.transmon_pocket.py."""
        component = transmon_pocket.TransmonPocket
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 4)
        self.assertEqual(metadata['short_name'], 'Pocket')
        self.assertEqual(metadata['_qgeometry_table_path'], 'True')
        self.assertEqual(metadata['_qgeometry_table_poly'], 'True')

    def test_qlibrary_transmon_cross_component_metadata(self):
        """Test component_metadata in qubits.transmon_cross.py."""
        component = transmon_cross.TransmonCross
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 3)
        self.assertEqual(metadata['short_name'], 'Cross')
        self.assertEqual(metadata['_qgeometry_table_poly'], 'True')
        self.assertEqual(metadata['_qgeometry_table_junction'], 'True')

    def test_qlibrary_star_qubit_component_metadata(self):
        """Test component_metadata in qubits.transmon_cross.py."""
        component = star_qubit.StarQubit
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 3)
        self.assertEqual(metadata['short_name'], 'Star')
        self.assertEqual(metadata['_qgeometry_table_poly'], 'True')
        self.assertEqual(metadata['_qgeometry_table_junction'], 'True')

    def test_qlibrary_transmon_cross_fl_component_metadata(self):
        """Test component_metadata in qubits.transmon_cross_fl.py."""
        component = transmon_cross_fl.TransmonCrossFL
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 3)
        self.assertEqual(metadata['short_name'], 'Q')
        self.assertEqual(metadata['_qgeometry_table_poly'], 'True')
        self.assertEqual(metadata['_qgeometry_table_path'], 'True')

    def test_qlibrary_transmon_pocket_6_component_metadata(self):
        """Test component_metadata in qubits.transmon_pcket_6.py."""
        component = transmon_pocket_6.TransmonPocket6
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 4)
        self.assertEqual(metadata['short_name'], 'Pocket')
        self.assertEqual(metadata['_qgeometry_table_poly'], 'True')
        self.assertEqual(metadata['_qgeometry_table_path'], 'True')
        self.assertEqual(metadata['_qgeometry_table_junction'], 'True')

    def test_qlibrary_transmon_pocket_teeth_component_metadata(self):
        """Test component_metadata in qubits.transmon_pcket_teeth.py."""
        component = TransmonPocketTeeth
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 4)
        self.assertEqual(metadata['short_name'], 'Pocket')
        self.assertEqual(metadata['_qgeometry_table_poly'], 'True')
        self.assertEqual(metadata['_qgeometry_table_path'], 'True')
        self.assertEqual(metadata['_qgeometry_table_junction'], 'True')

    def test_qlibrary_squid_loop_component_metadata(self):
        """Test component_metadata in qubits.squid_loop.py."""
        component = SQUID_LOOP
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 1)
        self.assertEqual(metadata['short_name'], 'component')

    def test_qlibrary_transmon_coupler_01_component_metadata(self):
        """Test component_metadata in qubits.tunable_coupler_01.py."""
        component = tunable_coupler_01.TunableCoupler01
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 4)
        self.assertEqual(metadata['short_name'], 'Pocket')
        self.assertEqual(metadata['_qgeometry_table_poly'], 'True')
        self.assertEqual(metadata['_qgeometry_table_path'], 'True')
        self.assertEqual(metadata['_qgeometry_table_junction'], 'True')

    def test_qlibrary_qubit_jj_dolan_component_metadata(self):
        """Test component_metadata in jj_dolan."""
        component = jj_dolan
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 1)
        self.assertEqual(metadata['short_name'], 'component')

    def test_qlibrary_qubit_jj_manhattan_component_metadata(self):
        """Test component_metadata in jj_dolan."""
        component = jj_manhattan
        metadata = component.component_metadata
        self.assertEqual(len(metadata), 1)
        self.assertEqual(metadata['short_name'], 'component')

    def test_qlibrary_qcomponent_get_pin_names(self):
        """Test getting all the pin names."""
        design = designs.DesignPlanar()
        my_q_component = QComponent(design, "my_name-get-pin-names", make=False)

        spiral_list_0 = self.generate_spiral_list(0, 0)
        spiral_list_1 = self.generate_spiral_list(1, 2)

        my_q_component.add_pin('pin1-name',
                               np.array(spiral_list_0.coords)[-2:], 0.1)
        my_q_component.add_pin('pin2-name',
                               np.array(spiral_list_1.coords)[-2:], 0.1)

        my_pin_names = list(my_q_component.pin_names)
        my_pin_names.sort()
        self.assertEqual(len(my_pin_names), 2)
        self.assertEqual(my_pin_names[0], 'pin1-name')
        self.assertEqual(my_pin_names[1], 'pin2-name')
        my_q_component.delete()

    def test_qlibrary_qcomponent_add_and_get_pin(self):
        """Test getting a pin by name."""
        design = designs.DesignPlanar()
        my_q_component = QComponent(design, "my_name-get-pin", make=False)

        spiral_list_0 = self.generate_spiral_list(0, 0)
        spiral_list_1 = self.generate_spiral_list(1, 2)

        my_q_component.add_pin('pin1-name',
                               np.array(spiral_list_0.coords)[-2:], 0.1)
        my_q_component.add_pin('pin2-name',
                               np.array(spiral_list_1.coords)[-2:], 0.1)

        my_pin = my_q_component.get_pin('pin2-name')
        self.assertEqual(len(my_pin), 10)
        self.assertEqual(len(my_pin['points']), 2)
        self.assertEqual(len(my_pin['points'][0]), 2)
        self.assertEqual(len(my_pin['points'][1]), 2)
        self.assertEqual(my_pin['points'][0][0], -34.)
        self.assertEqual(my_pin['points'][0][1], 32.)
        self.assertEqual(my_pin['points'][1][0], -34.)
        self.assertEqual(my_pin['points'][1][1], -33.)
        self.assertEqual(len(my_pin['middle']), 2)
        self.assertEqual(my_pin['middle'][0], -34.)
        self.assertEqual(my_pin['middle'][1], -0.5)
        self.assertEqual(len(my_pin['normal']), 2)
        self.assertEqual(my_pin['normal'][0], 1.)
        self.assertEqual(my_pin['normal'][1], -0.)
        self.assertEqual(len(my_pin['tangent']), 2)
        self.assertEqual(my_pin['tangent'][0], 0.)
        self.assertEqual(my_pin['tangent'][1], -1.)
        self.assertEqual(my_pin['width'], 65.0)
        self.assertEqual(my_pin['gap'], 0.06)
        self.assertEqual(my_pin['chip'], 'main')
        self.assertEqual(my_pin['parent_name'], 1)
        self.assertEqual(my_pin['net_id'], 0)
        self.assertEqual(my_pin['length'], 0)
        my_q_component.delete()

    def test_qlibrary_get_component_geometry_dict(self):
        """Test get_component_geometry_dict in qgeometries_handler.py."""
        design = designs.DesignPlanar()
        transmon_pocket.TransmonPocket(design, 'Q1')
        transmon_pocket.TransmonPocket(design, 'Q2')

        q1_actual = (-0.2275, 0.015, 0.2275, 0.105)
        q2_actual = (-0.2275, -0.105, 0.2275, -0.015)

        q1_dict = design._qgeometry.get_component_geometry_dict('Q1')
        q2_dict = design._qgeometry.get_component_geometry_dict('Q2')
        q1_result = q1_dict['poly'][0].bounds
        q2_result = q2_dict['poly'][1].bounds

        self.assertEqual(len(q1_actual), len(q1_result))
        self.assertEqual(len(q2_actual), len(q2_result))

        for x, _ in enumerate(q1_actual):
            self.assertEqual(_, q1_result[x])

        for x, _ in enumerate(q2_actual):
            self.assertEqual(_, q2_result[x])

    def test_qlibrary_get_component_geometry_list(self):
        """Test get_component_geometry_list in qgeometries_handler.py."""
        design = designs.DesignPlanar()
        transmon_pocket.TransmonPocket(design, 'Q1')

        expected = [(-0.2275, 0.015, 0.2275, 0.105),
                    (-0.2275, -0.105, 0.2275, -0.015),
                    (-0.325, -0.325, 0.325, 0.325), (0.0, -0.015, 0.0, 0.015)]
        actual = design._qgeometry.get_component_geometry_list('Q1')
        self.assertEqual(len(actual), len(expected))
        length = len(expected)
        for x in range(length):
            self.assertEqual(len(expected[x]), len(actual[x].bounds))
            sub_length = len(expected[x])
            for y in range(sub_length):
                self.assertEqual(actual[x].bounds[y], expected[x][y])

    def test_qlibrary_get_component_geometry(self):
        """Test get_component_geometry in qgeometries_handler.py."""
        design = designs.DesignPlanar()
        transmon_pocket.TransmonPocket(design, 'Q1')

        expected = [(-0.2275, 0.015, 0.2275, 0.105),
                    (-0.2275, -0.105, 0.2275, -0.015),
                    (-0.325, -0.325, 0.325, 0.325), (0.0, -0.015, 0.0, 0.015)]
        actual = design._qgeometry.get_component_geometry('Q1')

        self.assertEqual(len(expected), len(actual))
        length = len(expected)
        for x in range(length):
            self.assertEqual(len(expected[x]), len(actual[x].bounds))
            sub_length = len(expected[x])
            for y in range(sub_length):
                self.assertEqual(actual[x].bounds[y], expected[x][y])

    def test_qlibrary_rename_component(self):
        """Test rename_component in element_handler.py."""
        design = designs.DesignPlanar()
        transmon_pocket.TransmonPocket(design, 'Q1')

        component_id = design.components['Q1'].id
        design.rename_component(component_id, 'Q1_new_name')

        self.assertEqual(design.components.keys(), ['Q1_new_name'])

    def test_qlibrary_delete_component(self):
        """Test delete_component in element_handler.py."""
        design = designs.DesignPlanar()
        transmon_pocket.TransmonPocket(design, 'Q1')
        transmon_pocket.TransmonPocket(design, 'Q2')
        transmon_pocket.TransmonPocket(design, 'Q3')

        before_junction_list = design.qgeometry.tables['junction'][
            'component'].tolist()
        before_poly_list = design.qgeometry.tables['poly']['component'].tolist()

        component_id = design.components['Q2'].id
        design.qgeometry.delete_component_id(component_id)

        after_junction_list = design.qgeometry.tables['junction'][
            'component'].tolist()
        after_poly_list = design.qgeometry.tables['poly']['component'].tolist()

        self.assertTrue(component_id in before_junction_list)
        self.assertTrue(component_id in before_poly_list)
        self.assertFalse(component_id in after_junction_list)
        self.assertFalse(component_id in after_poly_list)

    def test_qlibrary_anchored_path_intersecting(self):
        """Test intersecting function in anchored_path.py"""
        self.assertTrue(
            anchored_path.intersecting(np.array([1, 1]), np.array([3, 3]),
                                       np.array([1, 3]), np.array([3, 1])))
        self.assertFalse(
            anchored_path.intersecting(np.array([1, 1]), np.array([3, 3]),
                                       np.array([5, 5]), np.array([7, 7])))

    @staticmethod
    def generate_spiral_list(x: int, y: int):
        """Helper function to generate a sprital list.

        Args:
            x (int): x-coordinate
            y (int): y-coordinate

        Returns:
            list: A list of points
        """
        spiral_list = []

        for step in range(3):
            point_value = 20 + step * 5
            spiral_list.append((-point_value, -point_value))
            spiral_list.append((point_value, -point_value))
            spiral_list.append((point_value, point_value))
            spiral_list.append((-point_value - (1 + 4), point_value))

        point_value = 20 + (step + 1) * 5
        spiral_list.append((-point_value, -point_value))
        spiral_list = draw.LineString(spiral_list)

        spiral_list = draw.rotate(spiral_list, 0, origin=(x, y))
        spiral_list = draw.translate(spiral_list, x, y)

        return spiral_list


if __name__ == '__main__':
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""This is a utility module used to handle qt error messages on slots and
etc."""

import inspect
import logging
import traceback
import types
from functools import wraps

from PySide6 import QtCore
from PySide6.QtCore import Slot

from ... import logger

__all__ = ['slot_catch_error', 'do_debug']

#######################################################################################
# Core handler


def _qt_message_handler(mode, context, message):
    """The message handler is a function that prints out debug messages,
    warnings, critical and fatal error messages. The Qt library (debug mode)
    contains hundreds of warning messages that are printed when internal errors
    (usually invalid function arguments) occur. Qt built in release mode also
    contains such warnings unless QT_NO_WARNING_OUTPUT and/or
    QT_NO_DEBUG_OUTPUT have been set during compilation. If you implement your
    own message handler, you get total control of these messages.

    The default message handler prints the message to the standard output under X11
    or to the debugger under Windows. If it is a fatal message, the application
    aborts immediately.

    For more info, see https://doc.qt.io/qt-5/qtglobal.html#qInstallMessageHandler

    Args:
        mode (QtCore mode): the mode
        context (context): the context
        message (str): the message
    """

    if message.startswith(
            'QSocketNotifier: Multiple socket notifiers for same socket'):
        pass  # Caused by running %gui qt multiple times
    else:
        if mode == QtCore.QtMsgType.QtInfoMsg:
            mode = 'INFO'
        elif mode == QtCore.QtMsgType.QtWarningMsg:
            mode = 'WARNING'
        elif mode == QtCore.QtMsgType.QtCriticalMsg:
            mode = 'CRITICAL'
        elif mode == QtCore.QtMsgType.QtFatalMsg:
            mode = 'FATAL'
        else:
            mode = 'DEBUG'
        # logger.log(
        #     getattr(logging, 'CRITICAL'), 'line: %d, func: %s(), file: %s' %
        #     (context.line, context.function, context.file) + '  %s: %s\n' %
        #     (mode, message))

        # Log basic message details
        base_message = f"{mode}: {message}"

        # Include context if available
        if context.file and context.function:
            base_message += (
                f" (File: {context.file}, Line: {context.line}, Function: {context.function})"
            )
        else:
            base_message += " (No context available from Qt)"

        # Capture Python traceback for additional details
        python_traceback = "".join(traceback.format_stack(limit=10))

        # Log the message with the Python traceback
        logger.log(
            getattr(logging, mode, logging.DEBUG),
            f"{base_message}\nPython Traceback (most recent call last):\n{python_traceback}"
        )


#######################################################################################
# Auxilary handlers - mostly for debug purposes


def do_debug(msg, name='info'):
    """Utility function used to print debug statements from PySide6 Socket
    calls A bit of a cludge.

    Args:
        msg (str): Message to print or log to user
        name (str): info wran, debug, etc.  Defaults to 'info'.
    """

    # if 0:
    #     # This just gives the qt main loop traceback. Not useful.
    #     callers = []
    #     for i in range(1, 20):
    #         try:
    #             stack = inspect.stack()[i]
    #             callers += [f'{stack.function}[{stack.lineno}]']
    #         except Exception as e:  # pylint: disable=broad-except
    #             print("Exception during do_debug exception handling: " +
    #                   e.__repr__())
    #     callers = reversed(callers)
    #     callers = '\n'.join(callers)
    #     msg = callers + "\n" + str(msg) + '\n'

    getattr(logger, name)(msg)


def slot_catch_error(*args, catch=Exception, on_exception_emit=None):
    """This is a decorator for Slots where an exception in user code is caught,
    printed and a optional qtSignal with signature qtSignal(Exception, str) is
    emitted when that happens.

    Based on:
        https://stackoverflow.com/questions/18740884/preventing-pyqt-to-silence-exceptions-occurring-in-slots

    Args:
        args (arguments):  any valid types for the Slot.
        catch (Exception):  Type of the exception to catch.  Defaults to Exception.
        on_exception_emit (str):  name of a qtSignal to be emitted.
    """

    if len(args) == 0 or isinstance(args[0], types.FunctionType):
        args = []

    @Slot(*args)
    def slot_decorator(func):

        @wraps(func)
        def wrapper(*args, **kwargs):  # pylint: disable=unused-argument

            try:
                func(*args)

            except catch as e:  # pylint: disable=invalid-name,broad-except

                message = traceback.format_exc()
                message += '\n\nERROR in call by Metal GUI (see traceback above)\n'\
                    + f"\n{' module   :':12s} {wrapper.__module__}" \
                    + f"\n{' function :':12s} {wrapper.__qualname__}" \
                    + f"\n{' err msg  :':12s} {e.__repr__()}"\
                    + f"\n{' args; kws:':12s} {args}; {kwargs}" \

                do_debug(message, name='error')

                if on_exception_emit is not None:
                    # args[0] is instance of bound signal
                    qt_signal = getattr(args[0], on_exception_emit)
                    qt_signal.emit(e, wrapper.__name__)

        return wrapper

    return slot_decorator



from PySide6.QtWidgets import (QLabel, QScrollArea, QVBoxLayout, QLabel,
                               QGroupBox)
from typing import List
from .build_history_scroll_area_ui import Ui_BuildHistory


class BuildHistoryScrollArea(QScrollArea, Ui_BuildHistory):
    """Widget for holding & displaying build logs of the component.

    This class extends the `QScrollArea` class
    """

    def __init__(self,
                 previous_builds: List[str],
                 parent=None,
                 *args,
                 **kwargs):
        """
        Args:
            previous_builds (List[str]):
                List of all previous attempted builds of the components
                (including successes and failures). Each attempted
                build is a string that will be displayed in this window.
                Assumes  previous_builds[0] is the latest log. For example. ['10PM error', '2PM error', '10AM error' ...]
                BuildHistoryScrollArea should always display the latest logs at the top
            parent (QWidget):
                Parent widget if necessary

        """
        super(BuildHistoryScrollArea, self).__init__(parent, *args, **kwargs)
        self.setupUi(self)
        self._previous_builds = previous_builds
        self._display_logs()

    def _display_logs(self):
        """Create UI for BuildHistoryScrollAreas."""
        for build_log_indx in range(len(self._previous_builds)):
            label = QLabel(self._previous_builds[build_log_indx])
            if 'ERROR' in self._previous_builds[build_log_indx]:
                label.setStyleSheet('color: red')
            self.build_display_vertical_layout.addWidget(label)
        self.adjustSize()
        self.show()



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""SQUID_loop.

.. code-block:
:      
    |                                                   |
    |                                                   |
    |-----seg a-----|JJ|-- seg b-------|                |
    |                                  |              plate 2
    |                                  |                |
  plate1                             seg c ---seg d-----|
    |                                  |                |
    |                                  |                |
    |--seg a lower--|JJ|--seg b lower--|                |
    |                                                   |
    |                                                   |
"""
from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.core.base import QComponent


class SQUID_LOOP(QComponent):
    """
    The base "SQUID_LOOP" inherits the "QComponent" class.

    This creates a simple SQUID loop consisting of two Josephson
    junctions (JJs) located on opposite sides of a rectangular loop.
    The design consists of eight separate rectangles: plate1, 
    segment a (and segment a lower), segment b (and segment b lower),
    segment c, segment d and plate 2. The two JJs are located between
    segments and a and b (and also between segments a lower and b
    lower.) 
    
    .. image::
        SQUID_LOOP.png

    .. meta::
        :description: Squid Loop

    Default Options:
        * plate1_width: '5.5um' -- width of plate1 (left)  
        * plate1_height: '40um' -- height of plate1 (left)
        * plate1_pos_x: '0' -- origin of the plate1 (left)
        * plate1_pos_y: '0' -- origin of the plate1 (left) 
        * squid_gap: '10um' -- space between 'seg a' and 'seg a lower'
        * segment_a_length: '10um' -- length of seg a
        * segment_a_width: '1um' -- width of seg a 
        * JJ_gap: '0.5um' -- space between seg a and seg b
        * segment_b_length: '5um' -- length of seg b
        * segment_b_width: '1um' -- width of seg b
        * segment_c_width: '10um' -- length of seg c
        * segment_d_length: '10um' -- length of seg d
        * segment_d_width: '2um' -- width of seg d 
        * plate2_width: '6um' -- width of plate 2 (right)
        * plate2_height: '30um' -- height of plate 2 (right)
    """
    # Default drawing options
    default_options = Dict(plate1_width='5.5um',
                           plate1_height='40um',
                           plate1_pos_x='0',
                           plate1_pos_y='0',
                           squid_gap='10um',
                           segment_a_length='10um',
                           segment_a_width='1um',
                           JJ_gap='0.5um',
                           segment_b_length='5um',
                           segment_b_width='1um',
                           segment_c_width='1um',
                           segment_d_length='10um',
                           segment_d_width='2um',
                           plate2_width='6um',
                           plate2_height='30um')
    """Default drawing options"""

    # Name prefix of component, if user doesn't provide name
    component_metadata = Dict(short_name='component')
    """Component metadata"""

    def make(self):
        """Convert self.options into QGeometry."""

        p = self.parse_options()  # Parse the string options into numbers

        # draw the lower pad as a rectangle
        plate1 = draw.rectangle(p.plate1_width, p.plate1_height, p.plate1_pos_x,
                                p.plate1_pos_y)

        segment_a = draw.rectangle(p.segment_a_length, p.segment_a_width,
                                   0.5 * (p.plate1_width + p.segment_a_length),
                                   0.5 * (p.squid_gap + p.segment_a_width))

        segment_a_lower = draw.translate(
            segment_a, 0.0, -1.0 * (p.squid_gap + p.segment_a_width))

        segment_b = draw.rectangle(
            p.segment_b_length, p.segment_b_width,
            0.5 * (p.plate1_width + p.segment_b_length) + p.JJ_gap +
            p.segment_a_length, 0.5 * (p.squid_gap + p.segment_b_width))

        segment_b_lower = draw.translate(
            segment_b, 0.0, -1.0 * (p.squid_gap + p.segment_b_width))

        segment_c = draw.rectangle(
            p.segment_c_width,
            p.squid_gap + p.segment_a_width + p.segment_b_width,
            0.5 * (p.plate1_width + p.segment_c_width) + p.segment_a_length +
            p.segment_b_length + p.JJ_gap, p.plate1_pos_y)

        segment_d = draw.rectangle(
            p.segment_d_length, p.segment_d_width,
            0.5 * (p.plate1_width + p.segment_d_length) + p.segment_a_length +
            p.segment_b_length + p.JJ_gap + p.segment_c_width, p.plate1_pos_y)

        plate2 = draw.rectangle(
            p.plate2_width, p.plate2_height, 0.5 *
            (p.plate1_width + p.plate2_width) + p.segment_a_length + p.JJ_gap +
            p.segment_b_length + p.segment_c_width + p.segment_d_length,
            p.plate1_pos_y)

        design1 = draw.union(plate1, segment_a, segment_a_lower, segment_b,
                             segment_b_lower, segment_c, segment_d, plate2)

        # now translate and rotate the final structure
        design1 = draw.rotate(design1, p.orientation, origin=(0, 0))
        design1 = draw.translate(design1, p.pos_x, p.pos_y)

        geom = {'design': design1}
        self.add_qgeometry('poly', geom, layer=p.layer, subtract=False)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=too-many-return-statements
# pylint: disable-msg=no-else-return
# pylint: disable-msg=inconsistent-return-statements
# pylint: disable-msg=attribute-defined-outside-init
# pylint: disable-msg=broad-except
# pylint: disable-msg=import-outside-toplevel
# pylint: disable-msg=relative-beyond-top-level
# pylint: disable-msg=import-error
# pylint: disable-msg=too-few-public-methods
# pylint: disable-msg=too-many-instance-attributes
# pylint: disable-msg=invalid-name
# pylint: disable-msg=no-name-in-module
"""
Parameter Entry Window for displaying parameters for QComponent instantiation from GUI's
QLibrary tab
"""

import copy
import importlib
import inspect
import os
import random
from collections import OrderedDict
from collections.abc import Callable
from inspect import signature
from pathlib import Path
from typing import TYPE_CHECKING, Union, Type

import numpy as np
from PySide6 import QtGui, QtWidgets
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDockWidget, QWidget
from PySide6.QtWidgets import (QMainWindow, QMessageBox)

from qiskit_metal import designs
from qiskit_metal.qlibrary.core import QComponent
from qiskit_metal.toolbox_python.attr_dict import Dict
from .model_view.tree_delegate_param_entry import ParamDelegate
from .model_view.tree_model_param_entry import TreeModelParamEntry, LeafNode, Node
from .parameter_entry_window_ui import Ui_MainWindow

if TYPE_CHECKING:
    from ...main_window import MetalGUI


class ParameterEntryWindow(QMainWindow):
    """Parameter entry window class"""

    def __init__(self,
                 qcomp_class: Type,
                 design: designs.DesignPlanar,
                 parent: QWidget = None,
                 gui: 'MetalGUI' = None):
        """
        Parameter Entry Widget when qcomponent is chosen from GUI's QLibrary
        Args:
            qcomp_class (Type): QComponent to be instantiated
            design (DesignPlanar): Current design being used
            parent (QWidget): Parent widget
            gui (MetalGUI): Metal GUI
        """

        super().__init__(parent)
        self.qcomp_class = qcomp_class
        self._design = design
        self._gui = gui
        self.setWindowTitle("New " + self.qcomp_class.__name__)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.param_dictionary = {}
        # set during generate_model_data as a backup in case tree need be
        # reloaded
        self.reset_param_dictionary = {}

        self.model = TreeModelParamEntry(self,
                                         self.ui.qcomponent_param_tree_view,
                                         design=self._design)

        self.ui.qcomponent_param_tree_view.setModel(self.model)
        self.ui.qcomponent_param_tree_view.setItemDelegate(ParamDelegate(self))

        self.statusBar().hide()

        # should be moved to qt designer
        self.ui.create_qcomp_button.clicked.connect(self.instantiate_qcomponent)
        self.ui.add_k_v_row_button.pressed.connect(self.add_k_v_row)
        self.ui.nest_dictionary_button.pressed.connect(self.add_k_dict_row)
        self.ui.remove_button.pressed.connect(self.delete_row)

    def setup_pew(self):
        """Setup pew"""
        self.generate_model_data()
        self._setup_help()
        self._setup_source()

    @property
    def qcomponent_file_path(self):
        """Get file path to qcomponent
        """
        component = self.qcomp_class
        module = inspect.getmodule(component)
        filepath = inspect.getfile(module)
        # TypeError
        return filepath

    # Exception Handling
    class QComponentParameterEntryExceptionDecorators():
        """
        All exceptions in QComponentParameterEntry should result in a pop-up window.
        This class contains the decorators that control exception handling for all functions
        in QComponentParameterEntry
        """

        @classmethod
        def entry_exception_pop_up_warning(cls, func: Callable):
            """
            Throws up critical QMessageBox with current exception in the event an exception is
            thrown by func

            Args:
                func (Callable): current function causing exceptions - should  be ONLY  qcpe instance methods
                    because decorators
                assumes arg[0] is a self who has a valid logger

            """

            def wrapper(*args, **kwargs):
                try:

                    return func(*args, **kwargs)
                # if anticipated Exception throw up error window
                except (Exception) as lqce:

                    #cls.log_error(args[0], lqce, func, args, kwargs)
                    args[0].error_pop_up = QMessageBox()

                    error_message = "In function, " + str(
                        func.__name__) + "\n" + str(
                            lqce.__class__.__name__) + ":\n" + str(lqce)

                    # modality set by critical, Don't set Title -- will NOT show
                    # up on MacOs¥
                    args[0].error_pop_up.critical(args[0], "", error_message)

            return wrapper

    def _setup_help(self):
        """Called when we need to set a new help"""

        component = self.qcomp_class
        if component is None:
            raise Exception("No Component found.")

        filepath = self.qcomponent_file_path

        imagefilename = Path(filepath.replace(".py", ".png")).name
        imagepath = Path(
            filepath
        ).parent.parent.parent / "_gui" / "_imgs" / "components" / imagefilename

        doc_class = self.format_docstr(inspect.getdoc(component))
        doc_init = self.format_docstr(inspect.getdoc(component.__init__))

        text = """<body style="color:white;">"""
        text += f'''
        <div class="h1">Summary:</div>
        <table class="table ComponentHeader">
            <tbody>
                <tr> <th>Class</th><td>{component.__name__}</td></tr>
                <tr> <th>Module</th><td>{component.__class__.__module__}</td></tr>
                <tr> <th>Path</th><td style="text-color=#BBBBBB;"> {filepath}</td></tr>
            </tbody>
        </table>
        '''

        if imagepath.is_file():
            text += f'''<img src="{str(imagepath)}"/>'''

        text += f'''
            <div class="h1">Class docstring:</div>
            {doc_class}
            <div class="h1">Init docstring:</div>
            {doc_init}
        '''
        text += "</body>"

        my_help = QtWidgets.QTextEdit()
        my_help.setReadOnly(True)
        my_help.setHtml(text)
        self.ui.tab_help.layout().addWidget(my_help)

    # pylint: disable-msg=no-self-use
    @QComponentParameterEntryExceptionDecorators.entry_exception_pop_up_warning
    def format_docstr(self, doc: Union[str, None]) -> str:
        """Format a docstring

        Args:
            doc (Union[str, None]): string to format

        Returns:
            str: formatted string
        """

        if doc is None:
            return ''
        doc = doc.strip()
        text = f"""
    <pre>
    <code class="DocString">{doc}</code>
    </pre>
        """
        return text

    def _setup_source(self):
        """Called when we need to set up a new source"""

        filepath = self.qcomponent_file_path

        text_source = QtWidgets.QTextEdit(self.ui.tab_source)
        text_source.setReadOnly(True)

        text_doc = QtGui.QTextDocument(text_source)
        try:  # For source doc
            #import pygments
            from pygments import highlight
            from pygments.formatters import HtmlFormatter
            from pygments.lexers import get_lexer_by_name
        except ImportError as e:
            self._design.logger.error(
                f'Error: Could not load python package \'pygments\'; Error: {e}'
            )
            highlight = None
            HtmlFormatter = None
            get_lexer_by_name = None

        text = Path(filepath).read_text()
        if highlight is None:
            text_doc.setPlainText(text)
        else:
            lexer = get_lexer_by_name("python", stripall=True)
            formatter = HtmlFormatter(linenos='inline')
            self._html_css_lex = formatter.get_style_defs('.highlight')
            text_doc.setDefaultStyleSheet(self._html_css_lex)
            text_html = highlight(text, lexer, formatter)
            text_doc.setHtml(text_html)
        text_source.moveCursor(QtGui.QTextCursor.Start)
        text_source.ensureCursorVisible()
        text_source.setDocument(text_doc)

        # set background colour to match that of formatter and text default to black
        wiget_background_html = '.highlight { background: '
        newstr = self._html_css_lex[self._html_css_lex.
                                    index(wiget_background_html) +
                                    len(wiget_background_html):]
        bg_color_1 = newstr[:newstr.find(';')]
        bg_color_2 = newstr[:newstr.find('}')]
        bg_color = bg_color_2 if len(bg_color_2) < len(
            bg_color_1) else bg_color_1
        text_source.setStyleSheet(f"""
        background-color: {bg_color}; 
        color: #000000;
        """)

        self.ui.tab_source.layout().addWidget(text_source)

    @QComponentParameterEntryExceptionDecorators.entry_exception_pop_up_warning
    def add_k_v_row(self):
        """ Add key, value row to parent row based on what row is highlighted in treeview"""
        cur_index = self.ui.qcomponent_param_tree_view.currentIndex()

        key = "fake-param"
        value = "value"
        self.model.add_new_leaf_node(cur_index, key, value)

    @QComponentParameterEntryExceptionDecorators.entry_exception_pop_up_warning
    def add_k_dict_row(self):
        """ Add key, dictionary-value to parent row based on what row is highlighted in treeview"""
        cur_index = self.ui.qcomponent_param_tree_view.currentIndex()

        fake_dict = "fake-dict"
        fakekey = "key"
        fakevalue = "value"
        self.model.add_new_branch_node(cur_index, fake_dict, fakekey, fakevalue)

    @QComponentParameterEntryExceptionDecorators.entry_exception_pop_up_warning
    def delete_row(self):
        """Delete highlight row"""
        cur_index = self.ui.qcomponent_param_tree_view.currentIndex()
        self.model.delete_node(cur_index)

    def generate_model_data(self):
        """
        Use QComponent's default_options and parameter (with typing) to populate Param Entry Window
        """

        param_dict = {}
        class_signature = signature(self.qcomp_class.__init__)

        for _, param in class_signature.parameters.items():
            if self.is_param_usable(param):
                if param.default:
                    param_dict[param.name] = param.default
                else:
                    class_name = self.qcomp_class.__name__ if param.name == 'name' else None
                    param_dict[param.name] = create_default_from_type(
                        param.annotation, param_name=class_name)

        ## Dealing with default options
        try:

            options = self.qcomp_class.get_template_options(self._design)

        except Exception as e:
            self._design.logger.warning(
                f"Could not use template_options for component: {e}")
            if 'default_options' in self.qcomp_class.__dict__:
                options = self.qcomp_class.default_options

        if options is not None:
            arg_options = 'options'

            # deepcopy so instance's options aren't pointing
            # to same dict as template_options nor default_options
            copied_options = copy.deepcopy(options)
            param_dict[arg_options] = copied_options

            # custom options for connection_pads
            def_con_pads = '_default_connection_pads'
            con_pads = 'connection_pads'
            if def_con_pads in param_dict[arg_options]:
                print(param_dict[arg_options])
                if con_pads not in param_dict[arg_options] or len(
                        param_dict[arg_options][con_pads]) < 1:
                    param_dict[arg_options][con_pads] = {
                        'a': param_dict[arg_options][def_con_pads]
                    }
                param_dict[arg_options].pop(def_con_pads)

        self.param_dictionary = param_dict
        self.reset_param_dictionary = copy.deepcopy(param_dict)
        self.model.init_load(param_dict)

    @staticmethod
    def is_param_usable(param):
        """Determines if a given parameter is usable."""
        ignore_params = {'self', 'design', 'make', 'kwargs', 'args'}
        if param.name in ignore_params:
            return False

        if_no_default_then_ignore_params = {'options_connection_pads'}
        if param.name in if_no_default_then_ignore_params:
            return param.default is not None

        return True

    @QComponentParameterEntryExceptionDecorators.entry_exception_pop_up_warning
    def instantiate_qcomponent(self):
        """Instantiate self.qcomp_class"""

        self.traverse_model_to_create_dictionary()

        self.qcomp_class(self._design, **self.current_dict)
        if self._gui is not None:  # for the sake of testing, we won't have gui
            self._gui.refresh()
            self._gui.autoscale()
        self.close()

    @QComponentParameterEntryExceptionDecorators.entry_exception_pop_up_warning
    def traverse_model_to_create_dictionary(self):
        """Traverse model to create parameter entry dictionary given to self.qcomp_class"""
        parameter_dict = {}
        r = self.model.root
        self.recursively_get_params(parameter_dict, r)

        self.current_dict = parameter_dict[""]

    @QComponentParameterEntryExceptionDecorators.entry_exception_pop_up_warning
    def recursively_get_params(self, parent_dict, cur_node):
        """Helper function for traverse_model_to_create_dictionary"""
        try:
            if isinstance(cur_node, LeafNode):
                parent_dict[cur_node.name] = cur_node.get_real_value()
                return

            c_dict = cur_node.get_empty_dictionary()
            parent_dict[cur_node.name] = c_dict
            for child in cur_node.children:

                self.recursively_get_params(c_dict, child[Node.NODE])
        except Exception as e:
            raise Exception(
                f"Unable to add node:{self.model.node_str(cur_node)} to {parent_dict} due to: {e}"
            ) from e


def create_parameter_entry_window(gui: 'MetalGUI',
                                  abs_file_path: str,
                                  parent=None) -> QtWidgets.QWidget:
    """Creates the spawned window that has the Parameter Entry Window
    """
    cur_class = get_class_from_abs_file_path(abs_file_path)
    if cur_class is None:
        gui.logger.error("Unable to get class from abs file: ", abs_file_path)
        return None

    # check for valid class
    if not issubclass(cur_class, QComponent):
        cur_class.error_pop_up = QMessageBox()

        error_message = f"The class {cur_class} is not a valid subclass of QComponent"

        # modality set by critical, Don't set Title -- will NOT show
        # up on MacOs¥
        cur_class.error_pop_up.critical(gui.main_window, "", error_message)
        return None

    if not parent:
        parent = gui.main_window  # gui.component_window.ui.tabHelp

    param_window = ParameterEntryWindow(cur_class, gui.design, parent, gui)

    param_window.dock = dockify(param_window, "New " + cur_class.__name__, gui)

    try:
        param_window.setup_pew()
    except Exception as e:
        param_window.dock.deleteLater()

        cur_class.error_pop_up = QMessageBox()

        error_message = f"Error while trying to instantiate object: {e}"

        # modality set by critical, Don't set Title -- will NOT show
        # up on MacOs¥
        cur_class.error_pop_up.critical(gui.main_window, "", error_message)
        return None

    param_window.dock.show()
    param_window.dock.raise_()
    param_window.dock.activateWindow()

    return param_window


def dockify(main_window, docked_title, gui):
    """Dockify the given GUI

    Args:
        gui (MetalGUI): the GUI

    Returns:
        QDockWidget: the widget
    """
    # Dockify
    main_window.dock_widget = QDockWidget(docked_title, gui.main_window)
    dock = main_window.dock_widget
    dock.setWidget(main_window)

    dock.setAllowedAreas(Qt.RightDockWidgetArea)
    dock.setFloating(True)
    dock.resize(1200, 700)

    gui.logger.info("dockified window: " + docked_title)
    return dock


def get_class_from_abs_file_path(abs_file_path: str):
    """
    Gets the corresponding class object for the absolute file path to the file containing that
    class definition

    Args:
        abs_file_path (str): absolute file path to the file containing the QComponent class definition

    getting class from absolute file path -
    https://stackoverflow.com/questions/452969/does-python-have-an-equivalent-to-java-class-forname

    """
    qis_abs_path = abs_file_path[abs_file_path.index(__name__.split('.')[0]):]

    # Windows users' qis_abs_path may use os.sep or '/' due to PySide's
    # handling of file names
    qis_mod_path = qis_abs_path.replace(os.sep, '.')[:-len('.py')]
    qis_mod_path = qis_mod_path.replace("/",
                                        '.')  # users cannot use '/' in filename

    cur_module = importlib.import_module(qis_mod_path)
    members = inspect.getmembers(cur_module, inspect.isclass)
    class_owner = qis_mod_path.split('.')[-1]
    for memtup in members:
        if len(memtup) > 1:
            if str(memtup[1].__module__).endswith(class_owner):
                return memtup[1]


def create_default_from_type(my_t: type, param_name: str = None):
    """
    Create default values for a given type.

    Args:
        my_t (type): argument type.
        param_name (str): default parameter name (will replace "fake-param") for non-dictionary
            types.

    Returns:
        object: a default parameter.
    """
    if param_name is not None:
        return param_name + "-" + str(random.randint(0, 1000))
    if my_t == int:
        return 0
    elif my_t == float:
        return 0.0
    elif my_t == str:
        return "fake-param-" + str(random.randint(0, 1000))
    elif my_t == bool:
        return True
    elif my_t == dict:
        return {
            "fake-param-" + str(random.randint(0, 1000)): "fake-param"
        }  # can't have empty branch nodes
    elif my_t == OrderedDict:
        return OrderedDict({0: "zeroth"})
    elif my_t == Dict:
        return Dict(falseparam1=Dict(falseparam2="false-param",
                                     falseparam3="false-param"))
    elif my_t is None:
        return "fake-param-" + str(random.randint(0, 1000))
    else:
        return np.ndarray(1)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Custom decorators for testing."""

from typing import Callable
from functools import wraps
import time


def timeout(timeout_seconds: int):
    """Simple decorator that fails a test that takes longer than time_seconds
    to run.

    Args:
        timeout_seconds (int): Maximum amount of time, in seconds, your test should run

    Returns:
        Returns nothing but does assert a failure if the test takes longer to run than
        timeout_seconds
    """

    def inner_timeout_decorater(test_function: Callable):

        @wraps(test_function)
        def timer_wrapper(self, *args, **kwargs):
            start = time.perf_counter()
            test_function(self, *args, **kwargs)
            end = time.perf_counter()
            self.assertLessEqual(
                end - start,
                timeout_seconds,
                msg=
                f"{test_function.__name__} ran in {end - start:0.4f} seconds but "
                + "needed to run sub {timeout_seconds}")

        return timer_wrapper

    return inner_timeout_decorater



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

#  This a launch structure used on BlueJayV2, used for wire bonding
#  There is no CPW tee attached to this p#

# Imports required for drawing

# import numpy as np # (currently not used, may be needed later for component customization)
from qiskit_metal import draw
from qiskit_metal.toolbox_python.attr_dict import Dict
from qiskit_metal.qlibrary.core import QComponent

# Define class and options for the launch geometry


class LaunchpadWirebond(QComponent):
    r"""Launch pad to feed/read signals to/from the chip.

    Inherits 'QComponent' class.

    .. image::
        LaunchpadWirebond.png

    .. meta::
        :description: Launchpad Wirebond

    Creates a 50 ohm launch pad with a ground pocket cutout.
    Limited but expandable parameters to control the launchpad polygons.
    The (0,0) point is the center of the necking of the launch tip.
    The pin attaches directly to the built in lead length at its midpoint

    Pocket and pad:
        Pocket and launch pad geometries are currently fixed.
        (0,0) point is the midpoint of the necking of the launch tip.
        Pocket is a negative shape that is cut out of the ground plane

    Values (unless noted) are strings with units included, (e.g., '30um')

    Sketch:
        Below is a sketch of the launch
        ::

            -----------
            |          \
            |      ---------\\
            |      |    0    |    (0,0) pin at midpoint of necking, before the lead
            |      ---------//
            |          /
            -----------

            y
            ^
            |
            |------> x

    .. image::
        LaunchpadWirebond.png

    Default Options:
        * trace_width: 'cpw_width' -- Width of the transmission line attached to the launch pad
        * trace_gap: 'cpw_gap' -- Gap of the transmission line
        * lead_length: '25um' -- Length of the transmission line attached to the launch pad
        * pad_width: '80um' -- Width of the launch pad
        * pad_height: '80um' -- Height of the launch pad
        * pad_gap: '58um' -- Gap of the launch pad
        * taper_height: '122um' -- Height of the taper from the launch pad to the transmission line
    """

    default_options = Dict(trace_width='cpw_width',
                           trace_gap='cpw_gap',
                           lead_length='25um',
                           pad_width='80um',
                           pad_height='80um',
                           pad_gap='58um',
                           taper_height='122um')
    """Default options"""

    TOOLTIP = """Launch pad to feed/read signals to/from the chip."""

    def make(self):
        """This is executed by the user to generate the qgeometry for the
        component."""

        p = self.p

        pad_width = p.pad_width
        pad_height = p.pad_height
        pad_gap = p.pad_gap
        trace_width = p.trace_width
        trace_width_half = trace_width / 2.
        pad_width_half = pad_width / 2.
        lead_length = p.lead_length
        taper_height = p.taper_height
        trace_gap = p.trace_gap

        pad_gap = p.pad_gap
        #########################################################

        # Geometry of main launch structure
        # The shape is a polygon and we prepare this point as orientation is 0 degree
        launch_pad = draw.Polygon([
            (0, trace_width_half), (-taper_height, pad_width_half),
            (-(pad_height + taper_height), pad_width_half),
            (-(pad_height + taper_height), -pad_width_half),
            (-taper_height, -pad_width_half), (0, -trace_width_half),
            (lead_length, -trace_width_half), (lead_length, trace_width_half),
            (0, trace_width_half)
        ])

        # Geometry pocket (gap)
        # Same way applied for pocket
        pocket = draw.Polygon([(0, trace_width_half + trace_gap),
                               (-taper_height, pad_width_half + pad_gap),
                               (-(pad_height + taper_height + pad_gap),
                                pad_width_half + pad_gap),
                               (-(pad_height + taper_height + pad_gap),
                                -(pad_width_half + pad_gap)),
                               (-taper_height, -(pad_width_half + pad_gap)),
                               (0, -(trace_width_half + trace_gap)),
                               (lead_length, -(trace_width_half + trace_gap)),
                               (lead_length, trace_width_half + trace_gap),
                               (0, trace_width_half + trace_gap)])

        # These variables are used to graphically locate the pin locations
        main_pin_line = draw.LineString([(lead_length, trace_width_half),
                                         (lead_length, -trace_width_half)])

        # Create polygon object list
        polys1 = [main_pin_line, launch_pad, pocket]

        # Rotates and translates all the objects as requested. Uses package functions in
        # 'draw_utility' for easy rotation/translation
        polys1 = draw.rotate(polys1, p.orientation, origin=(0, 0))
        polys1 = draw.translate(polys1, xoff=p.pos_x, yoff=p.pos_y)
        [main_pin_line, launch_pad, pocket] = polys1

        # Adds the object to the qgeometry table
        self.add_qgeometry('poly', dict(launch_pad=launch_pad), layer=p.layer)

        # Subtracts out ground plane on the layer its on
        self.add_qgeometry('poly',
                           dict(pocket=pocket),
                           subtract=True,
                           layer=p.layer)

        # Generates the pins
        self.add_pin('tie', main_pin_line.coords, trace_width)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# Tree model for Ansys HFSS renderer

import PySide6
from PySide6 import QtWidgets
from PySide6.QtWidgets import QTreeView, QWidget

from .widgets.bases.dict_tree_base import QTreeModel_Base

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .main_window import MetalGUI


class RendererHFSS_Model(QTreeModel_Base):
    """Tree model for Ansys HFSS renderer.

    Args:
        QTreeModel_Base (QAbstractItemModel): Base class for nested dicts
    """

    def __init__(self, parent: QWidget, gui: 'MetalGUI', view: QTreeView):
        """Editable table with drop-down rows for HFSS renderer options.
        Organized as a tree model where child nodes are more specific
        properties of a given parent node.

        Args:
            parent (QWidget): The parent widget
            gui (MetalGUI): The main user interface
            view (QTreeView): View corresponding to a tree structure
        """
        super().__init__(parent=parent,
                         gui=gui,
                         view=view,
                         child='HFSS renderer')

    @property
    def data_dict(self) -> dict:
        """Return a reference to the (nested) dictionary containing the
        data."""
        return self.design.renderers.hfss.options
